# -*- coding: utf-8 -*-

"""
Exceptions for CleverCSV

Author: Gertjan van den Burg

"""

from .cparser import Error as ParserError


class Error(ParserError):
    pass


class NoDetectionResult(Exception):
    pass



# -*- coding: utf-8 -*-

"""
Break ties in the data consistency measure.

Author: Gertjan van den Burg

"""

from typing import List
from typing import Optional

from .cparser_util import parse_string
from .dialect import SimpleDialect
from .utils import pairwise


def tie_breaker(
    data: str, dialects: List[SimpleDialect]
) -> Optional[SimpleDialect]:
    """
    Break ties between dialects.

    This function is used to break ties where possible between two, three, or
    four dialects that receive the same value for the data consistency measure.

    Parameters
    ----------
    data: str
        The data as a single string
    dialects: list
        Dialects that are tied

    Returns
    -------
    dialect: SimpleDialect
        One of the dialects from the list provided or None.


    """
    if len(dialects) == 2:
        return break_ties_two(data, dialects[0], dialects[1])
    elif len(dialects) == 3:
        return break_ties_three(data, dialects[0], dialects[1], dialects[2])
    elif len(dialects) == 4:
        return break_ties_four(data, dialects)
    return None


def reduce_pairwise(
    data: str, dialects: List[SimpleDialect]
) -> Optional[List[SimpleDialect]]:
    """Reduce the set of dialects by breaking pairwise ties

    Parameters
    ----------

    data: str
        The data of the file as a string

    dialects: list
        List of SimpleDialect objects

    Returns
    -------
    dialects: list
        List of SimpleDialect objects.

    """
    equal_delim = len(set([d.delimiter for d in dialects])) == 1
    if not equal_delim:
        return None  # TODO: This might be wrong, it can just return the input!

    # First, identify dialects that result in the same parsing result.
    equal_dialects = []
    for a, b in pairwise(dialects):
        X = list(parse_string(data, a))
        Y = list(parse_string(data, b))
        if X == Y:
            equal_dialects.append((a, b))

    # Try to break the ties in these pairs
    new_dialects = set()
    visited = set()
    for A, B in equal_dialects:
        ans = break_ties_two(data, A, B)
        if ans is not None:
            new_dialects.add(ans)
        visited.add(A)
        visited.add(B)

    # and add the dialects that we didn't visit
    for d in dialects:
        if d not in visited:
            new_dialects.add(d)

    return list(new_dialects)


def _dialects_only_differ_in_field(
    A: SimpleDialect, B: SimpleDialect, field: str
) -> bool:
    keys = ["delimiter", "quotechar", "escapechar"]
    return all(
        getattr(A, key) == getattr(B, key) for key in keys if key != field
    )


def break_ties_two(
    data: str, A: SimpleDialect, B: SimpleDialect
) -> Optional[SimpleDialect]:
    """Break ties between two dialects.

    This function breaks ties between two dialects that give the same score. We
    distinguish several cases:

    1. If delimiter and escapechar are the same and one of the quote characters
    is the empty string. We parse the file with both dialects and check if the
    parsing result is the same. If it is, the correct dialect is the one with
    no quotechar, otherwise it's the other one.
    2. If quotechar and escapechar are the same and the delimiters are comma
    and space, then we go for comma. Alternatively, if either of the delimiters
    is the hyphen, we assume it's the other dialect.
    3. If the delimiter and quotechar is the same and one dialect uses the
    escapchar and the other doesn't. We break this tie by checking if the
    escapechar has an effect and if it occurs an even or odd number of times.

    If it's none of these cases, we don't break the tie and return None.

    Parameters
    ----------

    data: str
        The data of the file as a string.

    A: SimpleDialect
        A potential dialect

    B: SimpleDialect
        A potential dialect

    Returns
    -------

    dialect: SimpleDialect or None
        The chosen dialect if the tie can be broken, None otherwise.

    """
    if _dialects_only_differ_in_field(A, B, "quotechar"):
        if A.quotechar == "" or B.quotechar == "":
            d_no = A if A.quotechar == "" else B
            d_yes = B if d_no == A else A

            X = list(parse_string(data, dialect=d_no))
            Y = list(parse_string(data, dialect=d_yes))

            if X == Y:
                # quotechar has no effect
                return d_no
            else:
                # quotechar has an effect
                return d_yes
    elif _dialects_only_differ_in_field(A, B, "delimiter"):
        if set([A.delimiter, B.delimiter]) == set([",", " "]):
            # Artifact due to type detection (comma as radix point)
            if A.delimiter == ",":
                return A
            else:
                return B
        elif A.delimiter == "-" or B.delimiter == "-":
            # Artifact due to type detection (dash as minus sign)
            if A.delimiter == "-":
                return B
            else:
                return A
    elif _dialects_only_differ_in_field(A, B, "escapechar"):
        Dnone, Descape = (A, B) if A.escapechar == "" else (B, A)

        X = list(parse_string(data, Dnone))
        Y = list(parse_string(data, Descape))

        # double check shape. Usually if the shape differs the pattern score
        # should have caught it, but if by a freakish occurance it hasn't then
        # we can't break this tie (for now)
        if len(X) != len(Y):
            return None
        for row_X, row_Y in zip(X, Y):
            if len(row_X) != len(row_Y):
                return None

        cells_escaped = []
        cells_unescaped = []
        for row_X, row_Y in zip(X, Y):
            for u, v in zip(row_X, row_Y):
                if u != v:
                    cells_unescaped.append(u)
                    cells_escaped.append(v)

        # We will break the ties in the following ways:
        #
        # If the escapechar precedes the quotechar an even number of times
        # within each offending cell, then we think it is a functional escape
        # and the escaped version is the correct dialect. Note that if an odd
        # number of escaped quotechars would occur, then the shape of the file
        # will be different if it is ignored. Only if it occurs an even number
        # of times within the cell can we get the same shape.
        for u in cells_unescaped:
            count = 0
            for a, b in pairwise(u):
                if a != Descape.escapechar:
                    continue
                if a == Descape.escapechar and b == Descape.quotechar:
                    count += 1
            if count > 0 and count % 2 == 0:
                return Descape
            else:
                return Dnone
    elif A.delimiter == B.delimiter:
        Aq, Ae = A.quotechar, A.escapechar
        Bq, Be = B.quotechar, B.escapechar
        if (Aq, Ae) == ("", "") or (Bq, Be) == ("", ""):
            # This case is activated if the escapechar+quotechar combination
            # occurs in the cells (i.e. "Jill\'s data") but no actual quoting
            # is done with the quote character.
            d_no = A if (Aq, Ae) == ("", "") else B
            d_yes = B if d_no == A else A

            X = list(parse_string(data, dialect=d_no))
            Y = list(parse_string(data, dialect=d_yes))

            if len(X) != len(Y):
                return None
            for row_X, row_Y in zip(X, Y):
                if len(row_X) != len(row_Y):
                    return None

            # if we're here, then there is no effect on structure.
            # we test if the only cells that differ are those that have an
            # escapechar+quotechar combination.
            assert isinstance(d_yes.escapechar, str)
            assert isinstance(d_yes.quotechar, str)
            eq = d_yes.escapechar + d_yes.quotechar
            for row_X, row_Y in zip(X, Y):
                for x, y in zip(row_X, row_Y):
                    if x != y:
                        if eq not in x:
                            return None

            # Now we know that the only cells that have the
            # escapechar+quotechar combination are the cause of the difference.
            # The right thing to do is to return the dialect that uses them.
            return d_yes

    return None


def break_ties_three(
    data: str, A: SimpleDialect, B: SimpleDialect, C: SimpleDialect
) -> Optional[SimpleDialect]:
    """Break ties between three dialects.

    If the delimiters and the escape characters are all equal, then we look for
    the dialect that has no quotechar. The tie is broken by calling
    :func:`break_ties_two` for the dialect without quotechar and another
    dialect that gives the same parsing result.

    If only the delimiter is the same for all dialects then use
    :func:`break_ties_two` on the dialects that do not have a quotechar,
    provided there are only two of these.

    Parameters
    ----------

    data: str
        The data of the file as a string

    A: SimpleDialect
        a dialect

    B: SimpleDialect
        a dialect

    C: SimpleDialect
        a dialect

    Returns
    -------

    dialect: Optional[SimpleDialect]
        The chosen dialect if the tie can be broken, None otherwise.

    Notes
    -----
    We have only observed one tie for each case during development, so
    this may need to be improved in the future.

    """

    equal_delim = A.delimiter == B.delimiter == C.delimiter
    equal_escape = A.escapechar == B.escapechar == C.escapechar

    if equal_delim and equal_escape:
        # difference is *only* in quotechar
        dialects = [A, B, C]

        pA = list(parse_string(data, A))
        pB = list(parse_string(data, B))
        pC = list(parse_string(data, C))

        if len(pA) != len(pB) or len(pA) != len(pC) or len(pB) != len(pC):
            return None

        p_none, d_none = next(
            (
                (p, d)
                for p, d in zip([pA, pB, pC], dialects)
                if d.quotechar == ""
            ),
            (None, None),
        )
        if p_none is None:
            return None
        assert d_none is not None

        rem = [
            (p, d) for p, d in zip([pA, pB, pC], dialects) if not p == p_none
        ]

        if len(rem) <= 1:
            # This case was reached for the file
            # 6da5ab459bcc7c3a5ed2e06d65810958.csv from the GitHub corpus of
            # the CSV paper. When fixing the delimiter to Tab, rem = [].
            # Try to reduce pairwise
            new_dialects = reduce_pairwise(data, dialects)
            if new_dialects is None:
                return None
            if len(new_dialects) == 1:
                return new_dialects[0]
            return None
        if p_none == rem[0][0]:
            return break_ties_two(data, d_none, rem[0][1])
        elif len(rem) > 1 and p_none == rem[1][0]:
            return break_ties_two(data, d_none, rem[1][1])
    elif equal_delim:
        # difference is in quotechar *and* escapechar

        # The reasoning here is as follows. If we are in this situation,
        # then there is both a potential escapechar and there are quotechars,
        # but the pattern score is the same and the type score can't make a
        # difference because no cells become clean if we interpret the
        # quote/escape correctly. This implies that the quote and escape do
        # have a function. Thus, we find the dialects that have a quote and
        # defer to break_ties_two.

        dialects = [A, B, C]
        with_quote = [d for d in dialects if d.quotechar != ""]

        if len(with_quote) != 2:
            return None

        return break_ties_two(data, with_quote[0], with_quote[1])

    return None


def break_ties_four(
    data: str, dialects: List[SimpleDialect]
) -> Optional[SimpleDialect]:
    """Break ties between four dialects.

    This function works by breaking the ties between pairs of dialects that
    result in the same parsing result (if any). If this reduces the number of
    dialects, then :func:`break_ties_three` or :func:`break_ties_two` is used,
    otherwise, the tie can't be broken.

    Ties are only broken if all dialects have the same delimiter.

    Parameters
    ----------

    data: str
        The data of the file as a string

    dialects: list
        List of SimpleDialect objects

    Returns
    -------
    dialect: Optional[SimpleDialect]
        The chosen dialect if the tie can be broken, None otherwise.

    Notes
    -----
    We have only observed one case during development where this
    function was needed. It may need to be revisited in the future if other
    examples are found.

    """
    # TODO: Check for length 4, handle more than 4 too?

    equal_delim = len(set([d.delimiter for d in dialects])) == 1
    if not equal_delim:
        return None

    reduced_dialects = reduce_pairwise(data, dialects)
    if reduced_dialects is None:
        return None

    # Defer to other functions if the number of dialects was reduced
    if len(reduced_dialects) == 1:
        return reduced_dialects[0]
    elif len(reduced_dialects) == 2:
        return break_ties_two(data, *reduced_dialects)
    elif len(reduced_dialects) == 3:
        return break_ties_three(data, *reduced_dialects)

    return None



# -*- coding: utf-8 -*-

"""
Unit tests for the type detection.

Author: Gertjan van den Burg

"""

import unittest

from typing import List

from clevercsv.detect_type import TypeDetector
from clevercsv.detect_type import type_score
from clevercsv.dialect import SimpleDialect


class TypeDetectorTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.td = TypeDetector()

    # NUMBERS

    def test_number(self) -> None:
        yes_number: List[str] = [
            "1",
            "2",
            "34",
            "56",
            "123",
            "789",
            "132.",
            "0.123",
            "0.10800212",
            "0.1231e-087",
            "10.789e09",
            "123.256e+08",
            "0.1231E-087",
            "10.789E09",
            "123.256E+08",
            "123,456,798.00",
            "23,456.798",
            "1,234.56",
            "1,123.",
            "1e5",
            "1.23e5",
            "-1",
            "-2",
            "-34",
            "-56",
            "-123",
            "-789",
            "-0.123",
            "-0.10800212",
            "-0.1231e-087",
            "-10.789e09",
            "-123.256e+08",
            "-0.1231E-087",
            "-10.789E09",
            "-123.256E+08",
            "-123,456,798.00",
            "-23,456.798",
            "-1,234.56",
            "+1",
            "+2",
            "+34",
            "+56",
            "+123",
            "+789",
            "+0.123",
            "+0.10800212",
            "+0.1231e-087",
            "+10.789e09",
            "+123.256e+08",
            "+0.1231E-087",
            "+10.789E09",
            "+123.256E+08",
            "+123,456,798.00",
            "+23,456.798",
            "+1,234.56",
            ".707",
            "-.707",
            "50,000.123",
            "1.000,123",
            "37.e88",
            "1.",
        ]
        for num in yes_number:
            with self.subTest(num=num):
                self.assertTrue(self.td.is_number(num))
        no_number: List[str] = [
            "0000.213654",
            "123.465.798",
            "0.5e0.5",
            "1,23.45",
            "12,34.56",
            "+00003",
            "0,132.6",
            "1,",
            "",
            "E14000537",
            "0e",
            ".",
            ",",
            "+E3",
            "1,",
        ]
        for num in no_number:
            with self.subTest(num=num):
                self.assertFalse(self.td.is_number(num))

    # DATES

    def test_date(self) -> None:
        yes_date: List[str] = [
            "031219",
            "03122019",
            "03-12-19",
            "03-12-2019",
            "03-5-19",
            "03-5-2019",
            "120319",
            "12032019",
            "12-03-19",
            "02-03-2019",
            "02-3-19",
            "02-3-2019",
            "19-12-3",
            "19-12-03",
            "19-2-3",
            "19-2-03",
            "8-21-19",
            "8-21-2019",
            "8-9-19",
            "8-9-2019",
            "7-12-19",
            "7-12-2019",
            "3-9-19",
            "3-9-2019",
            "191203",
            "20191121",
            "2019-12-3",
            "2019-12-21",
            "2019-3-9",
            "2019-3-21",
            "2019年11月21日",
            "2019年11月1日",
            "2019年3月21日",
            "2019年3月1日",
            "19年03月11日",
            "19年03月1日",
            "19年3月31日",
            "19年3月1日",
            "2019년11월21일",
            "2019년11월1일",
            "2019년3월21일",
            "2019년3월1일",
            "19년03월11일",
            "19년03월1일",
            "19년3월31일",
            "19년3월1일",
        ]
        for date in yes_date:
            with self.subTest(date=date):
                self.assertTrue(self.td.is_date(date))
        no_date: List[str] = [
            "2018|01|02",
            "30/07-88",
            "12.01-99",
            "5.024.2896",
            "2512-012.1",
            "12 01/2542",
        ]
        for date in no_date:
            with self.subTest(date=date):
                self.assertFalse(self.td.is_date(date))

    # DATETIME

    def test_datetime(self) -> None:
        yes_dt: List[str] = [
            "2019-01-12T04:01:23Z",
            "2021-09-26T12:13:31+01:00",
        ]
        for dt in yes_dt:
            with self.subTest(dt=dt):
                self.assertTrue(self.td.is_datetime(dt))
        no_date: List[str] = []
        for date in no_date:
            with self.subTest(date=date):
                self.assertFalse(self.td.is_datetime(dt))

    # URLs

    def test_url(self) -> None:
        # Some cases copied from https://mathiasbynens.be/demo/url-regex
        yes_url: List[str] = [
            "Cocoal.icio.us",
            "Websquash.com",
            "bbc.co.uk",
            "ebay.com",
            "en.wikipedia.com",
            "ftp://foo.bar/baz",
            "http://127.0.0.1",
            "http://127.0.0.1/uoshostel/web/app_dev.php/assets/img/size2.jpg",
            "http://1337.net",
            "http://142.42.1.1/",
            "http://142.42.1.1:8080/",
            "http://223.255.255.254",
            "http://a.b-c.de",
            "http://code.google.com/events/#&product=browser",
            "http://en.wikipedia.com",
            "http://experiment.local/frameworks/symphony2/web/app_dev.php/admin/categories",
            "http://foo.bar/?q=Test%20URL-encoded%20stuff",
            "http://foo.com/(something)?after=parens",
            "http://foo.com/blah_(wikipedia)#cite-1",
            "http://foo.com/blah_(wikipedia)_blah#cite-1",
            "http://foo.com/blah_blah",
            "http://foo.com/blah_blah/",
            "http://foo.com/blah_blah_(wikipedia)",
            "http://foo.com/blah_blah_(wikipedia)_(again)",
            "http://fridley-tigers.com",
            "http://gertjan.dev",
            "http://hi.fridley-tigers.com",
            "http://j.mp",
            "http://localhost/1234.html",
            "http://localhost/Symfony/web/app_dev.php/index",
            "http://localhost/pidev/WebSmartravel/web/app_dev.php/travel_admin/1/js/bootstrap.js",
            "http://localhost/webSmartravel/web/app_dev.php/admin",
            "http://mainhostel.localdev.com/app_dev.php/location",
            "http://simplegreensmoothies.com/Recipes/kiwi-strawberry-twist",
            "http://t.co/VSD0L81Yrt",
            "http://t.co/VSD0L81Yrt.html",
            "http://www.bbc.co.uk",
            "http://www.beloithistoricalsociety.com/hanchett.htm",
            "http://www.co-operativefood.co.uk/find-us/?q=UK&lat=52.451935&long=-1.887871&filters=Food&options=",
            "http://www.deutsche-wein-strasse.de/Panorama/Flemlingen/flemlingen.htm",
            "http://www.example.com/wpstyle/?p=364",
            "http://www.google.com/url?q=http%3A%2F%2Fbit.ly%2F1sneR2w&sa=D&sntz=1&usg=AFQjCNGs2NJSTFm8Dzx-755C0K0_KDuiew",
            "http://www.iceland.co.uk/",
            "http://www.iceland.co.uk/store-finder/",
            "http://www.iceland.co.uk/store-finder/?StoreFinderSearch=S45%209JE",
            "http://www.marksandspencer.com/MSStoreDetailsView?SAPStoreId=2804",
            "http://www.sainsburys.co.uk/sol/storelocator/storelocator_detail_view.jsp?storeId=4526&bmForm=store_details",
            "http://www.stackoverflow.com/",
            "https://en.wikipedia.com",
            "https://gertjan.dev",
            "https://google.com",
            "https://localhost",
            "https://www.example.com/foo/?bar=baz&inga=42&quux",
            "test.example.com/~name",
            "www.google.com",
            "www.google.com/",
            "www.menominee-nsn.gov/",
            "http://arxiv.org/abs/arXiv:1908.03213",
            "https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3474301/",
            "https://dl.acm.org/citation.cfm?id=3025626",
            "https://openreview.net/forum?id=S1x4ghC9tQ",
            "https://link.springer.com/article/10.1007/s10618-019-00631-5",
            "http://proceedings.mlr.press/v48/zhangf16.html",
            "https://papers.nips.cc/paper/7796-middle-out-decoding",
            "http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.89.6548",
            "http://localhost:81/test/web/app_dev.php/fr/intranet/admin/client/2",
        ]
        for url in yes_url:
            with self.subTest(url=url):
                self.assertTrue(self.td.is_url(url))

        no_url: List[str] = [
            "//",
            "///",
            "///a",
            "//a",
            ":// should fail",
            "a@b.com",
            "ftps://foo.bar/",
            "h://test",
            "http:// shouldfail.com",
            "http://",
            "http://#",
            "http://##",
            "http://##/",
            "http://-a.b.co",
            "http://-error-.invalid/",
            "http://.",
            "http://..",
            "http://../",
            "http://.www.foo.bar./",
            "http://.www.foo.bar/",
            "http:///a",
            "http://1.1.1.1.1",
            "http://123.123.123",
            "http://3628126748",
            "http://?",
            "http://??",
            "http://??/",
            "http://foo.bar/foo(bar)baz quux",
            "http://foo.bar?q=Spaces should be encoded",
            "http://www.foo.bar./",
            "rdar://1234",
        ]
        for url in no_url:
            with self.subTest(url=url):
                self.assertFalse(self.td.is_url(url))

    # Unicode_alphanum

    def test_unicode_alphanum(self) -> None:
        # These tests are by no means inclusive and ought to be extended in the
        # future.

        yes_alphanum: List[str] = ["this is a cell", "1231 pounds"]
        for unicode_alphanum in yes_alphanum:
            with self.subTest(unicode_alphanum=unicode_alphanum):
                self.assertTrue(self.td.is_unicode_alphanum(unicode_alphanum))
                self.assertTrue(
                    self.td.is_unicode_alphanum(
                        unicode_alphanum, is_quoted=True
                    )
                )

        no_alphanum: List[str] = ["https://www.gertjan.dev"]
        for unicode_alpanum in no_alphanum:
            with self.subTest(unicode_alpanum=unicode_alpanum):
                self.assertFalse(self.td.is_unicode_alphanum(unicode_alpanum))
                self.assertFalse(
                    self.td.is_unicode_alphanum(
                        unicode_alpanum, is_quoted=True
                    )
                )

        only_quoted: List[str] = ["this string, with a comma"]
        for unicode_alpanum in only_quoted:
            with self.subTest(unicode_alpanum=unicode_alpanum):
                self.assertFalse(
                    self.td.is_unicode_alphanum(
                        unicode_alpanum,
                    )
                )
                self.assertTrue(
                    self.td.is_unicode_alphanum(
                        unicode_alpanum, is_quoted=True
                    )
                )

    def test_bytearray(self) -> None:
        yes_bytearray: List[str] = [
            "bytearray(b'')",
            "bytearray(b'abc,*&@\"')",
            "bytearray(b'bytearray(b'')')",
        ]
        no_bytearray: List[str] = [
            "bytearray(b'abc",
            "bytearray(b'abc'",
            "bytearray('abc')",
            "abc,bytearray(b'def')",
        ]

        for case in yes_bytearray:
            with self.subTest(case=case):
                self.assertTrue(self.td.is_bytearray(case))

        for case in no_bytearray:
            with self.subTest(case=case):
                self.assertFalse(self.td.is_bytearray(case))

    # Unix path

    def test_unix_path(self) -> None:
        yes_path: List[str] = [
            "/Users/person/abc/def-ghi/blabla.csv.test",
            "/home/username/share/a/_b/c_d/e.py",
            "/home/username/share",
            "/home/username",
            "/home/username/",
            "~/share/",
            "./share",
        ]
        for path in yes_path:
            with self.subTest(path=path):
                self.assertTrue(self.td.is_unix_path(path))
        no_path = ["", "~share", ".share"]
        for path in no_path:
            with self.subTest(path=path):
                self.assertFalse(self.td.is_unix_path(path))

    """
    Type Score tests
    """

    def test_type_score_1(self) -> None:
        # theta_1 from paper
        cells = [
            ["7", "5; Mon", " Jan 12;6", "40"],
            ["100; Fri", " Mar 21;8", "23"],
            ["8", "2; Thu", " Sep 17; 2", "71"],
            ["538", "0;;7", "26"],
            ['"NA"; Wed', " Oct 4;6", "93"],
        ]
        data = "\n".join([",".join(x) for x in cells])
        dialect = SimpleDialect(delimiter=",", quotechar="", escapechar="")
        out = type_score(data, dialect)
        exp = 8 / 17
        self.assertAlmostEqual(exp, out)

    def test_type_score_2(self) -> None:
        # theta_2 from paper
        cells = [
            ["7,5", " Mon, Jan 12", "6,40"],
            ["100", " Fri, Mar 21", "8,23"],
            ["8,2", " Thu, Sep 17", "2,71"],
            ["538,0", "", "7,26"],
            ['"N/A"', " Wed, Oct 4", "6,93"],
        ]
        data = "\r\n".join([";".join(x) for x in cells])
        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        out = type_score(data, dialect)
        exp = 10 / 15
        self.assertAlmostEqual(exp, out)

    def test_type_score_3(self) -> None:
        # theta_3 from paper
        cells = [
            ["7,5", " Mon, Jan 12", "6,40"],
            ["100", " Fri, Mar 21", "8,23"],
            ["8,2", " Thu, Sep 17", "2,71"],
            ["538,0", "", "7,26"],
            ["N/A", " Wed, Oct 4", "6,93"],
        ]
        data = "\r".join([";".join(x) for x in cells])
        dialect = SimpleDialect(delimiter=";", quotechar='"', escapechar="")
        out = type_score(data, dialect)
        exp = 11 / 15
        self.assertAlmostEqual(exp, out)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

"""
DictReader and DictWriter.

This code is entirely copied from the Python csv module. The only exception is
that it uses the `reader` and `writer` classes from our package.

Author: Gertjan van den Burg

"""

from __future__ import annotations

import warnings

from collections import OrderedDict
from collections.abc import Collection

from typing import TYPE_CHECKING
from typing import Any
from typing import Generic
from typing import Iterable
from typing import Iterator
from typing import Literal
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import TypeVar
from typing import Union
from typing import cast

from clevercsv.read import reader
from clevercsv.write import writer

if TYPE_CHECKING:
    from clevercsv._types import SupportsWrite
    from clevercsv._types import _DialectLike
    from clevercsv._types import _DictReadMapping

_T = TypeVar("_T")


class DictReader(
    Generic[_T], Iterator["_DictReadMapping[Union[_T, Any], Union[str, Any]]"]
):
    def __init__(
        self,
        f: Iterable[str],
        fieldnames: Optional[Sequence[_T]] = None,
        restkey: Optional[str] = None,
        restval: Optional[str] = None,
        dialect: "_DialectLike" = "excel",
        *args: Any,
        **kwds: Any,
    ) -> None:
        self._fieldnames = fieldnames
        self.restkey = restkey
        self.restval = restval
        self.reader: reader = reader(f, dialect, *args, **kwds)
        self.dialect = dialect
        self.line_num = 0

    def __iter__(self) -> "DictReader[_T]":
        return self

    @property
    def fieldnames(self) -> Sequence[_T]:
        if self._fieldnames is None:
            try:
                fieldnames = next(self.reader)
                self._fieldnames = [cast(_T, f) for f in fieldnames]
            except StopIteration:
                pass

        assert self._fieldnames is not None

        # Note: this was added because I don't think it's expected that Python
        # simply drops information if there are duplicate headers. There is
        # discussion on this issue in the Python bug tracker here:
        # https://bugs.python.org/issue17537 (see linked thread therein). A
        # warning is easy enough to suppress and should ensure that the user
        # is at least aware of this behavior.
        if not len(self._fieldnames) == len(set(self._fieldnames)):
            warnings.warn(
                "fieldnames are not unique, some columns will be dropped."
            )

        self.line_num = self.reader.line_num
        return self._fieldnames

    @fieldnames.setter
    def fieldnames(self, value: Sequence[_T]) -> None:
        self._fieldnames = value

    def __next__(self) -> "_DictReadMapping[Union[_T, Any], Union[str, Any]]":
        if self.line_num == 0:
            self.fieldnames
        row = next(self.reader)
        self.line_num = self.reader.line_num

        while row == []:
            row = next(self.reader)

        d: _DictReadMapping = OrderedDict(zip(self.fieldnames, row))
        lf = len(self.fieldnames)
        lr = len(row)
        if lf < lr:
            d[self.restkey] = row[lf:]
        elif lf > lr:
            for key in self.fieldnames[lr:]:
                d[key] = self.restval
        return d


class DictWriter(Generic[_T]):
    def __init__(
        self,
        f: SupportsWrite[str],
        fieldnames: Collection[_T],
        restval: Optional[Any] = "",
        extrasaction: Literal["raise", "ignore"] = "raise",
        dialect: "_DialectLike" = "excel",
        *args: Any,
        **kwds: Any,
    ):
        self.fieldnames = fieldnames
        self.restval = restval
        if extrasaction.lower() not in ("raise", "ignore"):
            raise ValueError(
                "extrasaction (%s) must be 'raise' or 'ignore'" % extrasaction
            )
        self.extrasaction = extrasaction
        self.writer = writer(f, dialect, *args, **kwds)

    def writeheader(self) -> Any:
        header = dict(zip(self.fieldnames, self.fieldnames))
        return self.writerow(header)

    def _dict_to_list(self, rowdict: Mapping[_T, Any]) -> Iterator[Any]:
        if self.extrasaction == "raise":
            wrong_fields = rowdict.keys() - self.fieldnames
            if wrong_fields:
                raise ValueError(
                    "dict contains fields not in fieldnames: "
                    + ", ".join([repr(x) for x in wrong_fields])
                )
        return (rowdict.get(key, self.restval) for key in self.fieldnames)

    def writerow(self, rowdict: Mapping[_T, Any]) -> Any:
        return self.writer.writerow(self._dict_to_list(rowdict))

    def writerows(self, rowdicts: Iterable[Mapping[_T, Any]]) -> None:
        return self.writer.writerows(map(self._dict_to_list, rowdicts))



# -*- coding: utf-8 -*-

import json
import sys
import time

from typing import Any
from typing import Dict

from wilderness import Command

from clevercsv.wrappers import detect_dialect

from ._docs import FLAG_DESCRIPTIONS
from ._utils import parse_int


class DetectCommand(Command):
    _description = "Detect the dialect of a CSV file."

    def __init__(self) -> None:
        super().__init__(
            name="detect",
            title="Detect the dialect of a CSV file",
            description=self._description,
            extra_sections={"CleverCSV": "Part of the CleverCSV suite"},
        )

    def register(self) -> None:
        self.add_argument("path", help="Path to the CSV file")
        self.add_argument(
            "-c",
            "--consistency",
            action="store_true",
            help="Only use the consistency measure for detection.",
            description=(
                "By default, the dialect of CSV files is detected using "
                "atwo-step process. First, a strict set of checks is used to "
                "see if the file adheres to a very basic format (for example, "
                "when all cells in the file are integers). If none of these "
                "checks succeed, the data consistency measure of Van den "
                "Burg, et al. (2019) is used to detect the dialect. With this "
                "option, you can force the detection to always use the data "
                "consistency measure. This can be useful for testing or "
                "research purposes, for instance."
            ),
        )
        self.add_argument(
            "-e",
            "--encoding",
            help="Set the encoding of the file",
            description=FLAG_DESCRIPTIONS["encoding"],
        )
        self.add_argument(
            "-n",
            "--num-chars",
            help="Number of characters to use for detection",
            type=int,
            description=FLAG_DESCRIPTIONS["num-chars"],
        )
        group = self.add_mutually_exclusive_group()
        group.add_argument(
            "-p",
            "--plain",
            action="store_true",
            help="Print the components of the dialect on separate lines",
        )
        group.add_argument(
            "-j",
            "--json",
            action="store_true",
            help="Print the components of the dialect as a JSON object",
            description=(
                "Print the dialect to standard output in the form of a JSON "
                "object. This object will always have the 'delimiter', "
                "'quotechar', 'escapechar', and 'strict' keys. If "
                "--add-runtime is specified, it will also have a 'runtime' "
                "key."
            ),
        )
        self.add_argument(
            "--no-skip",
            action="store_true",
            help="Don't skip type detection for dialects with a low pattern score",
            description=(
                "The data consistency score used for dialect detection "
                "consists of two components: a pattern score and a type "
                "score. The type score lies between 0 and 1. When computing "
                "the data consistency measures for different dialects, we "
                "skip the computation of the type score if we see that the "
                "pattern score is lower than the best data consistency score "
                "we've seen so far. This option can be used to disable this "
                "behaviour and compute the type score for all dialects. This "
                "is mainly useful for debugging and testing purposes."
            ),
        )
        self.add_argument(
            "--add-runtime",
            action="store_true",
            help="Add the runtime of the detection to the detection output.",
        )

    def handle(self) -> int:
        verbose = self.args.verbose
        num_chars = parse_int(self.args.num_chars, "num-chars")
        method = "consistency" if self.args.consistency else "auto"
        skip = not self.args.no_skip

        t_start = time.time()
        dialect = detect_dialect(
            self.args.path,
            num_chars=num_chars,
            encoding=self.args.encoding,
            verbose=verbose,
            method=method,
            skip=skip,
        )
        runtime = time.time() - t_start

        if dialect is None:
            print("Error: Dialect detection failed.", file=sys.stderr)
            return 1

        if self.args.plain:
            print(f"delimiter = {dialect.delimiter}".strip())
            print(f"quotechar = {dialect.quotechar}".strip())
            print(f"escapechar = {dialect.escapechar}".strip())
            if self.args.add_runtime:
                print(f"runtime = {runtime}")
        elif self.args.json:
            dialect_dict: Dict[str, Any] = dialect.to_dict()
            if self.args.add_runtime:
                dialect_dict["runtime"] = runtime
            print(json.dumps(dialect_dict))
        else:
            print("Detected: " + str(dialect))
            if self.args.add_runtime:
                print(f"Runtime: {runtime:.6f} seconds")
        return 0



# -*- coding: utf-8 -*-

from csv import QUOTE_ALL
from csv import QUOTE_MINIMAL
from csv import QUOTE_NONE
from csv import QUOTE_NONNUMERIC

from .__version__ import __version__
from .cparser_util import field_size_limit
from .detect import Detector
from .detect import Detector as Sniffer
from .dialect import excel
from .dialect import excel_tab
from .dialect import unix_dialect
from .dict_read_write import DictReader
from .dict_read_write import DictWriter
from .exceptions import Error
from .read import reader
from .wrappers import detect_dialect
from .wrappers import read_dataframe
from .wrappers import read_dicts
from .wrappers import read_table
from .wrappers import stream_dicts
from .wrappers import stream_table
from .wrappers import write_table
from .write import writer

__all__ = [
    "QUOTE_ALL",
    "QUOTE_MINIMAL",
    "QUOTE_NONE",
    "QUOTE_NONNUMERIC",
    "__version__",
    "field_size_limit",
    "Detector",
    "Sniffer",
    "excel",
    "excel_tab",
    "unix_dialect",
    "DictReader",
    "DictWriter",
    "Error",
    "reader",
    "detect_dialect",
    "read_dataframe",
    "read_dicts",
    "read_table",
    "stream_dicts",
    "stream_table",
    "write_table",
    "writer",
]



# -*- coding: utf-8 -*-

from .code import CodeCommand
from .detect import DetectCommand
from .explore import ExploreCommand
from .standardize import StandardizeCommand
from .view import ViewCommand

__all__ = [
    "CodeCommand",
    "DetectCommand",
    "ExploreCommand",
    "StandardizeCommand",
    "ViewCommand",
]



# -*- coding: utf-8 -*-

"""
Unit tests for the dialect detection.

Author: Gertjan van den Burg

"""

import unittest

from clevercsv.detect import Detector


class DetectorTestCase(unittest.TestCase):
    # Initially we copy the results from CPython test suite.

    sample1 = """\
Harry's, Arlington Heights, IL, 2/1/03, Kimi Hayes
Shark City, Glendale Heights, IL, 12/28/02, Prezence
Tommy's Place, Blue Island, IL, 12/28/02, Blue Sunday/White Crow
Stonecutters Seafood and Chop House, Lemont, IL, 12/19/02, Week Back
"""
    sample2 = """\
'Harry''s':'Arlington Heights':'IL':'2/1/03':'Kimi Hayes'
'Shark City':'Glendale Heights':'IL':'12/28/02':'Prezence'
'Tommy''s Place':'Blue Island':'IL':'12/28/02':'Blue Sunday/White Crow'
'Stonecutters ''Seafood'' and Chop House':'Lemont':'IL':'12/19/02':'Week Back'
"""
    header1 = """\
"venue","city","state","date","performers"
"""
    sample3 = """\
05/05/03?05/05/03?05/05/03?05/05/03?05/05/03?05/05/03
05/05/03?05/05/03?05/05/03?05/05/03?05/05/03?05/05/03
05/05/03?05/05/03?05/05/03?05/05/03?05/05/03?05/05/03
"""

    sample4 = """\
2147483648;43.0e12;17;abc;def
147483648;43.0e2;17;abc;def
47483648;43.0;170;abc;def
"""

    sample5 = "aaa\tbbb\r\nAAA\t\r\nBBB\t\r\n"
    sample6 = "a|b|c\r\nd|e|f\r\n"
    sample7 = "'a'|'b'|'c'\r\n'd'|e|f\r\n"

    header2 = """\
"venue"+"city"+"state"+"date"+"performers"
"""
    sample8 = """\
Harry's+ Arlington Heights+ IL+ 2/1/03+ Kimi Hayes
Shark City+ Glendale Heights+ IL+ 12/28/02+ Prezence
Tommy's Place+ Blue Island+ IL+ 12/28/02+ Blue Sunday/White Crow
Stonecutters Seafood and Chop House+ Lemont+ IL+ 12/19/02+ Week Back
"""
    # adapted to be not broken
    sample9 = """\
'Harry''s'+ 'Arlington Heights'+ 'IL'+ '2/1/03'+ 'Kimi Hayes'
'Shark City'+ 'Glendale Heights'+' IL'+ '12/28/02'+ 'Prezence'
'Tommy''s Place'+ 'Blue Island'+ 'IL'+ '12/28/02'+ 'Blue Sunday/White Crow'
'Stonecutters ''Seafood'' and Chop House'+ 'Lemont'+ 'IL'+ '12/19/02'+ 'Week Back'
"""

    sample10 = """\
bytearray(b'fake data'),20:53:06,2019-09-01T19:28:21
bytearray(b'fake data'),19:33:15,2005-02-15T19:10:31
bytearray(b'fake data'),10:43:05,1992-10-12T14:49:24
bytearray(b'fake data'),10:36:49,1999-07-18T17:27:55
bytearray(b'fake data'),03:33:35,1982-04-24T17:38:45
bytearray(b'fake data'),14:49:47,1983-01-05T22:17:42
bytearray(b'fake data'),10:35:30,2006-10-27T02:30:45
"""

    sample11 = """\
"{""fake"": ""json"", ""fake2"":""json2""}",13:31:38,06:00:04+01:00
"{""fake"": ""json"", ""fake2"":""json2""}",22:13:29,14:20:11+02:00
"{""fake"": ""json"", ""fake2"":""json2""}",04:37:27,22:04:28+03:00
"{""fake"": ""json"", ""fake2"":""json2""}",04:25:28,23:12:53+01:00
"{""fake"": ""json"", ""fake2"":""json2""}",21:04:15,08:23:58+02:00
"{""fake"": ""json"", ""fake2"":""json2""}",10:37:03,11:06:42+05:30
"{""fake"": ""json"", ""fake2"":""json2""}",10:17:24,23:38:47+06:00
"{""fake"": ""json"", ""fake2"":""json2""}",00:02:51,20:04:45-06:00
"""

    def test_detect(self) -> None:
        # Adapted from CPython
        detector = Detector()
        dialect = detector.detect(self.sample1)
        assert dialect is not None
        self.assertEqual(dialect.delimiter, ",")
        self.assertEqual(dialect.quotechar, "")
        self.assertEqual(dialect.escapechar, "")

        dialect = detector.detect(self.sample2)
        assert dialect is not None
        self.assertEqual(dialect.delimiter, ":")
        self.assertEqual(dialect.quotechar, "'")
        self.assertEqual(dialect.escapechar, "")

    def test_delimiters(self) -> None:
        # Adapted from CPython
        detector = Detector()
        dialect = detector.detect(self.sample3)
        assert dialect is not None
        self.assertIn(dialect.delimiter, self.sample3)
        dialect = detector.detect(self.sample3, delimiters="?,")
        assert dialect is not None
        self.assertEqual(dialect.delimiter, "?")
        dialect = detector.detect(self.sample3, delimiters="/,")
        assert dialect is not None
        self.assertEqual(dialect.delimiter, "/")
        dialect = detector.detect(self.sample4)
        assert dialect is not None
        self.assertEqual(dialect.delimiter, ";")
        dialect = detector.detect(self.sample5)
        assert dialect is not None
        self.assertEqual(dialect.delimiter, "\t")
        dialect = detector.detect(self.sample6)
        assert dialect is not None
        self.assertEqual(dialect.delimiter, "|")
        dialect = detector.detect(self.sample7)
        assert dialect is not None
        self.assertEqual(dialect.delimiter, "|")
        self.assertEqual(dialect.quotechar, "'")
        dialect = detector.detect(self.sample8)
        assert dialect is not None
        self.assertEqual(dialect.delimiter, "+")
        dialect = detector.detect(self.sample9)
        assert dialect is not None
        self.assertEqual(dialect.delimiter, "+")
        self.assertEqual(dialect.quotechar, "'")
        dialect = detector.detect(self.sample10)
        assert dialect is not None
        self.assertEqual(dialect.delimiter, ",")
        self.assertEqual(dialect.quotechar, "")
        dialect = detector.detect(self.sample11)
        assert dialect is not None
        self.assertEqual(dialect.delimiter, ",")
        self.assertEqual(dialect.quotechar, '"')

    def test_has_header(self) -> None:
        detector = Detector()
        self.assertEqual(detector.has_header(self.sample1), False)
        self.assertEqual(
            detector.has_header(self.header1 + self.sample1), True
        )

    def test_has_header_regex_special_delimiter(self) -> None:
        detector = Detector()
        self.assertEqual(detector.has_header(self.sample8), False)
        self.assertEqual(
            detector.has_header(self.header2 + self.sample8), True
        )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

"""
Python utility functions that wrap the C parser.

"""

import io

from typing import Any
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from .cparser import Error as ParserError
from .cparser import Parser
from .dialect import SimpleDialect
from .exceptions import Error

_FIELD_SIZE_LIMIT: int = 128 * 1024


def field_size_limit(*args: Any, **kwargs: Any) -> int:
    """Get/Set the limit to the field size.

    This function is adapted from the one in the Python CSV module. See the
    documentation there.
    """
    global _FIELD_SIZE_LIMIT
    old_limit = _FIELD_SIZE_LIMIT
    all_args = list(args) + list(kwargs.values())
    if not 0 <= len(all_args) <= 1:
        raise TypeError(
            "field_size_limit expected at most 1 arguments, got %i"
            % len(all_args)
        )
    if len(all_args) == 0:
        return old_limit
    limit = all_args[0]
    if not isinstance(limit, int):
        raise TypeError("limit must be an integer")
    _FIELD_SIZE_LIMIT = int(limit)
    return old_limit


def _parse_data(
    data: Iterable[str],
    delimiter: str,
    quotechar: str,
    escapechar: str,
    strict: bool,
    return_quoted: bool = False,
) -> Iterator[Union[List[str], List[Tuple[str, bool]]]]:
    parser = Parser(
        data,
        delimiter=delimiter,
        quotechar=quotechar,
        escapechar=escapechar,
        field_limit=field_size_limit(),
        strict=strict,
        return_quoted=return_quoted,
    )
    try:
        for row in parser:
            yield row
    except ParserError as e:
        raise Error(str(e))


def parse_data(
    data: Iterable[str],
    dialect: Optional[SimpleDialect] = None,
    delimiter: Optional[str] = None,
    quotechar: Optional[str] = None,
    escapechar: Optional[str] = None,
    strict: Optional[bool] = None,
    return_quoted: bool = False,
) -> Iterator[Union[List[str], List[Tuple[str, bool]]]]:
    """Parse the data given a dialect using the C parser

    Parameters
    ----------
    data : iterable
        The data of the CSV file as an iterable

    dialect : SimpleDialect
        The dialect to use for the parsing. If None, the dialect with each
        component set to the empty string is used.

    delimiter : str
        The delimiter to use. If not None, overwrites the delimiter in the
        dialect.

    quotechar : str
        The quote character to use. If not None, overwrites the quote character
        in the dialect.

    escapechar : str
        The escape character to use. If not None, overwrites the escape
        character in the dialect.

    strict : bool
        Enable strict mode or not. If not None, overwrites the strict mode set
        in the dialect.

    return_quoted : bool
        For each cell, return a tuple "(field, is_quoted)" where the second
        element indicates whether the cell was a quoted cell or not.

    Yields
    ------
    rows : list
        The rows of the file as a list of cells.

    Raises
    ------
    Error : clevercsv.exceptions.Error
        When an error occurs during parsing.

    """
    if dialect is None:
        dialect = SimpleDialect("", "", "")

    delimiter_ = delimiter if delimiter is not None else dialect.delimiter
    quotechar_ = quotechar if quotechar is not None else dialect.quotechar
    escapechar_ = escapechar if escapechar is not None else dialect.escapechar
    strict_ = strict if strict is not None else dialect.strict

    yield from _parse_data(
        data,
        delimiter_,
        quotechar_,
        escapechar_,
        strict_,
        return_quoted=return_quoted,
    )


def parse_string(
    data: str,
    dialect: SimpleDialect,
    return_quoted: bool = False,
) -> Iterator[Union[List[str], List[Tuple[str, bool]]]]:
    """Utility for when the CSV file is encoded as a single string"""
    return parse_data(
        iter(io.StringIO(data, newline="")),
        dialect=dialect,
        return_quoted=return_quoted,
    )



# -*- coding: utf-8 -*-

"""
Unit tests for making abstractions

"""

import gzip
import json
import unittest

from pathlib import Path

from typing import Any
from typing import Dict
from typing import List

from clevercsv.cabstraction import c_merge_with_quotechar
from clevercsv.detect_pattern import base_abstraction
from clevercsv.detect_pattern import fill_empties
from clevercsv.detect_pattern import strip_trailing


class AbstractionTestCase(unittest.TestCase):
    def setUp(self) -> None:
        here = Path(__file__)
        this_dir = here.parent
        data_dir = this_dir / "data"
        testcases_file = data_dir / "abstraction_testcases.json.gz"
        if not testcases_file.exists():
            self._cases = []
        else:
            self._cases = self._load_cases(testcases_file)

    @staticmethod
    def _load_cases(filename: Path) -> List[Dict[str, Any]]:
        cases = []
        with gzip.open(filename, "rt", newline="", encoding="utf-8") as fp:
            for line in fp:
                cases.append(json.loads(line))
        return cases

    def test_abstraction_multi(self) -> None:
        if not self._cases:
            self.skipTest("no abstraction test cases found")

        for case in self._cases:
            content = case["content"]
            dialect = case["dialect"]

            exp_base = case["base_abstraction"]
            exp_merge = case["after_merge_with_quotechar"]
            exp_empties = case["after_fill_empties"]
            exp_trailing = case["after_strip_trailing"]
            with self.subTest(name=case["name"], kind="base"):
                base = base_abstraction(
                    content,
                    dialect["delimiter"],
                    dialect["quotechar"],
                    dialect["escapechar"],
                )
                self.assertEqual(base, exp_base)

            with self.subTest(name=case["name"], kind="merge"):
                merge = c_merge_with_quotechar(base)
                self.assertEqual(merge, exp_merge)

            with self.subTest(name=case["name"], kind="empties"):
                empties = fill_empties(merge)
                self.assertEqual(empties, exp_empties)

            with self.subTest(name=case["name"], kind="trailing"):
                trailing = strip_trailing(empties)
                self.assertEqual(trailing, exp_trailing)


if __name__ == "__main__":
    unittest.main()



from typing import Any

__ALL__: Any
VERSION: Any
ATTRIBUTES: Any
HIGHLIGHTS: Any
COLORS: Any
RESET: str

def colored(
    text,
    color: Any | None = ...,
    on_color: Any | None = ...,
    attrs: Any | None = ...,
): ...
def cprint(
    text,
    color: Any | None = ...,
    on_color: Any | None = ...,
    attrs: Any | None = ...,
    **kwargs
) -> None: ...



# -*- coding: utf-8 -*-

"""
Unit tests for the DictReader and DictWriter classes

Most of these are the same as in CPython, but we also test the cases where
CleverCSV's behavior differs.

"""

import io
import tempfile
import unittest

from typing import Any
from typing import Dict

import clevercsv

from clevercsv.dict_read_write import DictReader


class DictTestCase(unittest.TestCase):
    ############################
    # Start tests from CPython #

    def test_writeheader_return_value(self) -> None:
        with tempfile.TemporaryFile("w+", newline="") as fp:
            writer = clevercsv.DictWriter(fp, fieldnames=["f1", "f2", "f3"])
            writeheader_return_value = writer.writeheader()
            self.assertEqual(writeheader_return_value, 10)

    def test_write_simple_dict(self) -> None:
        with tempfile.TemporaryFile("w+", newline="") as fp:
            writer = clevercsv.DictWriter(fp, fieldnames=["f1", "f2", "f3"])
            writer.writeheader()
            fp.seek(0)
            self.assertEqual(fp.readline(), "f1,f2,f3\r\n")
            writer.writerow({"f1": 10, "f3": "abc"})
            fp.seek(0)
            fp.readline()  # header
            self.assertEqual(fp.read(), "10,,abc\r\n")

    def test_write_multiple_dict_rows(self) -> None:
        fp = io.StringIO()
        writer = clevercsv.DictWriter(fp, fieldnames=["f1", "f2", "f3"])
        writer.writeheader()
        self.assertEqual(fp.getvalue(), "f1,f2,f3\r\n")
        writer.writerows(
            [
                {"f1": 1, "f2": "abc", "f3": "f"},
                {"f1": 2, "f2": 5, "f3": "xyz"},
            ]
        )
        self.assertEqual(fp.getvalue(), "f1,f2,f3\r\n1,abc,f\r\n2,5,xyz\r\n")

    def test_write_no_fields(self) -> None:
        fp = io.StringIO()
        self.assertRaises(TypeError, clevercsv.DictWriter, fp)

    def test_write_fields_not_in_fieldnames(self) -> None:
        with tempfile.TemporaryFile("w+", newline="") as fp:
            writer = clevercsv.DictWriter(fp, fieldnames=["f1", "f2", "f3"])
            # Of special note is the non-string key (CPython issue 19449)
            content: Dict[Any, Any] = {"f4": 10, "f2": "spam", 1: "abc"}
            with self.assertRaises(ValueError) as cx:
                writer.writerow(content)
            exception = str(cx.exception)
            self.assertIn("fieldnames", exception)
            self.assertIn("'f4'", exception)
            self.assertNotIn("'f2'", exception)
            self.assertIn("1", exception)

    def test_typo_in_extrasaction_raises_error(self) -> None:
        fp = io.StringIO()
        self.assertRaises(
            ValueError,
            clevercsv.DictWriter,
            fp,
            ["f1", "f2"],
            extrasaction="raised",
        )

    def test_write_field_not_in_field_names_raise(self) -> None:
        fp = io.StringIO()
        writer = clevercsv.DictWriter(fp, ["f1", "f2"], extrasaction="raise")
        dictrow = {"f0": 0, "f1": 1, "f2": 2, "f3": 3}
        self.assertRaises(
            ValueError, clevercsv.DictWriter.writerow, writer, dictrow
        )

    def test_write_field_not_in_field_names_ignore(self) -> None:
        fp = io.StringIO()
        writer = clevercsv.DictWriter(fp, ["f1", "f2"], extrasaction="ignore")
        dictrow = {"f0": 0, "f1": 1, "f2": 2, "f3": 3}
        clevercsv.DictWriter.writerow(writer, dictrow)
        self.assertEqual(fp.getvalue(), "1,2\r\n")

    def test_read_dict_fields(self) -> None:
        with tempfile.TemporaryFile("w+") as fp:
            fp.write("1,2,abc\r\n")
            fp.seek(0)
            reader: DictReader[str] = clevercsv.DictReader(
                fp, fieldnames=["f1", "f2", "f3"]
            )
            self.assertEqual(next(reader), {"f1": "1", "f2": "2", "f3": "abc"})

    def test_read_dict_no_fieldnames(self) -> None:
        with tempfile.TemporaryFile("w+") as fp:
            fp.write("f1,f2,f3\r\n1,2,abc\r\n")
            fp.seek(0)
            reader: DictReader[str] = clevercsv.DictReader(fp)
            self.assertEqual(next(reader), {"f1": "1", "f2": "2", "f3": "abc"})
            self.assertEqual(reader.fieldnames, ["f1", "f2", "f3"])

    # Two test cases to make sure existing ways of implicitly setting
    # fieldnames continue to work.  Both arise from discussion in issue3436.
    def test_read_dict_fieldnames_from_file(self) -> None:
        with tempfile.TemporaryFile("w+") as fp:
            fp.write("f1,f2,f3\r\n1,2,abc\r\n")
            fp.seek(0)
            reader: DictReader[str] = clevercsv.DictReader(
                fp, fieldnames=next(clevercsv.reader(fp))
            )
            self.assertEqual(reader.fieldnames, ["f1", "f2", "f3"])
            self.assertEqual(next(reader), {"f1": "1", "f2": "2", "f3": "abc"})

    def test_read_dict_fieldnames_chain(self) -> None:
        import itertools

        with tempfile.TemporaryFile("w+") as fp:
            fp.write("f1,f2,f3\r\n1,2,abc\r\n")
            fp.seek(0)
            reader: DictReader[str] = clevercsv.DictReader(fp)
            first = next(reader)
            for row in itertools.chain([first], reader):
                self.assertEqual(reader.fieldnames, ["f1", "f2", "f3"])
                self.assertEqual(row, {"f1": "1", "f2": "2", "f3": "abc"})

    def test_read_long(self) -> None:
        with tempfile.TemporaryFile("w+") as fp:
            fp.write("1,2,abc,4,5,6\r\n")
            fp.seek(0)
            reader = clevercsv.DictReader(fp, fieldnames=["f1", "f2"])
            self.assertEqual(
                next(reader),
                {"f1": "1", "f2": "2", None: ["abc", "4", "5", "6"]},
            )

    def test_read_long_with_rest(self) -> None:
        with tempfile.TemporaryFile("w+") as fp:
            fp.write("1,2,abc,4,5,6\r\n")
            fp.seek(0)
            reader = clevercsv.DictReader(
                fp, fieldnames=["f1", "f2"], restkey="_rest"
            )
            self.assertEqual(
                next(reader),
                {"f1": "1", "f2": "2", "_rest": ["abc", "4", "5", "6"]},
            )

    def test_read_long_with_rest_no_fieldnames(self) -> None:
        with tempfile.TemporaryFile("w+") as fp:
            fp.write("f1,f2\r\n1,2,abc,4,5,6\r\n")
            fp.seek(0)
            reader: DictReader[str] = clevercsv.DictReader(fp, restkey="_rest")
            self.assertEqual(reader.fieldnames, ["f1", "f2"])
            self.assertEqual(
                next(reader),
                {"f1": "1", "f2": "2", "_rest": ["abc", "4", "5", "6"]},
            )

    def test_read_short(self) -> None:
        with tempfile.TemporaryFile("w+") as fp:
            fp.write("1,2,abc,4,5,6\r\n1,2,abc\r\n")
            fp.seek(0)
            reader = clevercsv.DictReader(
                fp, fieldnames="1 2 3 4 5 6".split(), restval="DEFAULT"
            )
            self.assertEqual(
                next(reader),
                {"1": "1", "2": "2", "3": "abc", "4": "4", "5": "5", "6": "6"},
            )
            self.assertEqual(
                next(reader),
                {
                    "1": "1",
                    "2": "2",
                    "3": "abc",
                    "4": "DEFAULT",
                    "5": "DEFAULT",
                    "6": "DEFAULT",
                },
            )

    def test_read_multi(self) -> None:
        sample = [
            "2147483648,43.0e12,17,abc,def\r\n",
            "147483648,43.0e2,17,abc,def\r\n",
            "47483648,43.0,170,abc,def\r\n",
        ]

        reader = clevercsv.DictReader(
            sample, fieldnames="i1 float i2 s1 s2".split()
        )
        self.assertEqual(
            next(reader),
            {
                "i1": "2147483648",
                "float": "43.0e12",
                "i2": "17",
                "s1": "abc",
                "s2": "def",
            },
        )

    def test_read_with_blanks(self) -> None:
        reader = clevercsv.DictReader(
            ["1,2,abc,4,5,6\r\n", "\r\n", "1,2,abc,4,5,6\r\n"],
            fieldnames="1 2 3 4 5 6".split(),
        )
        self.assertEqual(
            next(reader),
            {"1": "1", "2": "2", "3": "abc", "4": "4", "5": "5", "6": "6"},
        )
        self.assertEqual(
            next(reader),
            {"1": "1", "2": "2", "3": "abc", "4": "4", "5": "5", "6": "6"},
        )

    def test_read_semi_sep(self) -> None:
        reader = clevercsv.DictReader(
            ["1;2;abc;4;5;6\r\n"],
            fieldnames="1 2 3 4 5 6".split(),
            delimiter=";",
        )
        self.assertEqual(
            next(reader),
            {"1": "1", "2": "2", "3": "abc", "4": "4", "5": "5", "6": "6"},
        )

    # End tests from CPython #
    ##########################

    ###################################
    # Start tests added for CleverCSV #

    def test_read_duplicate_fieldnames(self) -> None:
        reader: DictReader[str] = clevercsv.DictReader(
            ["f1,f2,f1\r\n", "a", "b", "c"]
        )
        with self.assertWarns(UserWarning):
            reader.fieldnames

    # End tests added for CleverCSV #
    #################################


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

"""
Detect the dialect using the data consistency measure.

Author: Gertjan van den Burg

"""

from dataclasses import dataclass
from functools import lru_cache

from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional

from . import field_size_limit
from .break_ties import tie_breaker
from .cparser_util import parse_string
from .detect_pattern import pattern_score
from .detect_type import DEFAULT_EPS_TYPE
from .detect_type import TypeDetector
from .dialect import SimpleDialect
from .potential_dialects import get_dialects


@dataclass
class ConsistencyScore:
    """Container to track the consistency score calculation

    Parameters
    ----------
    P : float
        The pattern score

    T : Optional[float]
        The type score. Can be None if not computed for speed.

    Q : Optional[float]
        The consistency score. Can be None if not computed for speed.

    """

    P: float
    T: Optional[float]
    Q: Optional[float]


class ConsistencyDetector:
    """Detect the dialect with the data consistency measure

    This class uses the data consistency measure to detect the dialect. See the
    paper for details.

    Parameters
    ----------
    skip : bool
        Skip computation of the type score for dialects with a low pattern
        score.

    verbose : bool
        Print out the dialects considered and their scores.

    cache_capacity: int
        The size of the cache for type detection. Caching the type detection
        result greatly speeds up the computation of the consistency measure.
        The size of the cache can be changed to trade off memory use and speed.

    """

    def __init__(
        self,
        skip: bool = True,
        verbose: bool = False,
        cache_capacity: int = 100_000,
    ) -> None:
        self._skip = skip
        self._verbose = verbose
        self._type_detector = TypeDetector()
        self._cache_capacity = cache_capacity

        # NOTE: A bit ugly but allows setting the cache size dynamically
        @lru_cache(cache_capacity)
        def cached_is_known_type(cell: str, is_quoted: bool) -> bool:
            return self._type_detector.is_known_type(cell, is_quoted)

        self._cached_is_known_type = cached_is_known_type

    def detect(
        self, data: str, delimiters: Optional[List[str]] = None
    ) -> Optional[SimpleDialect]:
        """Detect the dialect using the consistency measure

        Parameters
        ----------
        data : str
            The data of the file as a string

        delimiters : iterable
            List of delimiters to consider. If None, the :func:`get_delimiters`
            function is used to automatically detect this (as described in the
            paper).

        Returns
        -------
        dialect : SimpleDialect
            The detected dialect. If no dialect could be detected, returns None.

        """
        self._cached_is_known_type.cache_clear()

        # TODO: probably some optimization there too
        dialects = get_dialects(data, delimiters=delimiters)

        # TODO: This is not thread-safe and this object can simply own a Parser
        # for each dialect and set the limit directly there (we can also cache
        # the best parsing result)
        old_limit = field_size_limit(len(data) + 1)

        scores = self.compute_consistency_scores(data, dialects)
        best_dialects = ConsistencyDetector.get_best_dialects(scores)
        result: Optional[SimpleDialect] = None
        if len(best_dialects) == 1:
            result = best_dialects[0]
        else:
            result = tie_breaker(data, best_dialects)

        field_size_limit(old_limit)
        return result

    def compute_consistency_scores(
        self, data: str, dialects: List[SimpleDialect]
    ) -> Dict[SimpleDialect, ConsistencyScore]:
        """Compute the consistency score for each dialect

        This function computes the consistency score for each dialect. This is
        done by first computing the pattern score for a dialect. If the class
        is instantiated with ``skip`` set to False, it also computes the type
        score for each dialect. If ``skip`` is True (the default), the type
        score is only computed if the pattern score is larger or equal to the
        current best combined score.

        Parameters
        ----------
        data : str
            The data of the file as a string

        dialects : Iterable[SimpleDialect]
            An iterable of delimiters to consider.

        Returns
        -------
        scores : Dict[SimpleDialect, ConsistencyScore]
            A map with a :class:`ConsistencyScore` object for each dialect
            provided as input.

        """

        scores: Dict[SimpleDialect, ConsistencyScore] = {}
        incumbent_score = -float("inf")
        for dialect in sorted(dialects):
            P = pattern_score(data, dialect)
            if P < incumbent_score and self._skip:
                scores[dialect] = ConsistencyScore(P, None, None)
                if self._verbose:
                    print("%15r:\tP = %15.6f\tskip." % (dialect, P))
                continue

            T = self.compute_type_score(data, dialect)
            Q = P * T
            incumbent_score = max(incumbent_score, Q)
            scores[dialect] = ConsistencyScore(P, T, Q)
            if self._verbose:
                print(
                    "%15r:\tP = %15.6f\tT = %15.6f\tQ = %15.6f"
                    % (dialect, P, T, Q)
                )
        return scores

    @staticmethod
    def get_best_dialects(
        scores: Dict[SimpleDialect, ConsistencyScore]
    ) -> List[SimpleDialect]:
        """Identify the dialects with the highest consistency score"""
        Qscores = [score.Q for score in scores.values()]
        Qmax = -float("inf")
        for q in Qscores:
            if q is None:
                continue
            Qmax = max(Qmax, q)
        return [d for d, score in scores.items() if score.Q == Qmax]

    def compute_type_score(
        self, data: str, dialect: SimpleDialect, eps: float = DEFAULT_EPS_TYPE
    ) -> float:
        """Compute the type score"""
        total = known = 0
        for row in parse_string(data, dialect, return_quoted=True):
            assert all(isinstance(cell, tuple) for cell in row)
            for cell, is_quoted in row:
                total += 1
                known += self._cached_is_known_type(cell, is_quoted=is_quoted)
        if not total:
            return eps
        return max(eps, known / total)


def detect_dialect_consistency(
    data: str,
    delimiters: Optional[Iterable[str]] = None,
    skip: bool = True,
    verbose: bool = False,
) -> Optional[SimpleDialect]:
    """Helper function that wraps ConsistencyDetector"""
    # Mostly kept for backwards compatibility
    consistency_detector = ConsistencyDetector(skip=skip, verbose=verbose)
    if delimiters is not None:
        delimiters = list(delimiters)
    return consistency_detector.detect(data, delimiters=delimiters)



# -*- coding: utf-8 -*-

"""
Unit tests based on fuzzing

"""

import unittest

import clevercsv


class FuzzingTestCase(unittest.TestCase):
    def test_sniffer_fuzzing(self) -> None:
        strings = ['"""', "```", "\"'", "'@'", "'\"", "'''", "O##P~`"]
        for string in strings:
            with self.subTest(string=string):
                try:
                    _ = clevercsv.Sniffer().sniff(string)
                except clevercsv.exceptions.Error:
                    pass



# -*- coding: utf-8 -*-

"""
Unit tests for the CCSV write module.

Author: Gertjan van den Burg

"""


import csv
import tempfile
import unittest

from typing import Any
from typing import Iterable

import clevercsv

from clevercsv.dialect import SimpleDialect


class WriterTestCase(unittest.TestCase):
    def _write_test(
        self, fields: Iterable[Any], expect: str, **kwargs: Any
    ) -> None:
        with tempfile.TemporaryFile("w+", newline="", prefix="ccsv_") as fp:
            writer = clevercsv.writer(fp, **kwargs)
            writer.writerow(fields)
            fp.seek(0)
            self.assertEqual(fp.read(), expect + writer.dialect.lineterminator)

    def _write_error_test(
        self, exc: type[Exception], fields: Any, **kwargs: Any
    ) -> None:
        with tempfile.TemporaryFile("w+", newline="", prefix="ccsv_") as fp:
            writer = clevercsv.writer(fp, **kwargs)
            with self.assertRaises(exc):
                writer.writerow(fields)
            fp.seek(0)
            self.assertEqual(fp.read(), "")

    def test_write_arg_valid(self) -> None:
        self._write_error_test(clevercsv.Error, None)
        self._write_test((), "")
        self._write_test([None], '""')
        self._write_error_test(
            clevercsv.Error, [None], quoting=clevercsv.QUOTE_NONE
        )

        # Check that exceptions are passed up the chain
        class BadList:
            def __len__(self) -> int:
                return 10

            def __getitem__(self, i: int) -> None:
                if i > 2:
                    raise OSError

        self._write_error_test(OSError, BadList())

        class BadItem:
            def __str__(self) -> str:
                raise OSError

        self._write_error_test(OSError, [BadItem()])

    def test_write_bigfield(self) -> None:
        bigstring = "X" * 50000
        self._write_test(
            [bigstring, bigstring], "%s,%s" % (bigstring, bigstring)
        )

    def test_write_quoting(self) -> None:
        self._write_test(["a", 1, "p,q"], 'a,1,"p,q"')
        self._write_error_test(
            clevercsv.Error, ["a", 1, "p,q"], quoting=clevercsv.QUOTE_NONE
        )
        self._write_test(
            ["a", 1, "p,q"], 'a,1,"p,q"', quoting=clevercsv.QUOTE_MINIMAL
        )
        self._write_test(
            ["a", 1, "p,q"], '"a",1,"p,q"', quoting=clevercsv.QUOTE_NONNUMERIC
        )
        self._write_test(
            ["a", 1, "p,q"], '"a","1","p,q"', quoting=clevercsv.QUOTE_ALL
        )
        self._write_test(
            ["a\nb", 1], '"a\nb","1"', quoting=clevercsv.QUOTE_ALL
        )

    def test_write_simpledialect(self) -> None:
        self._write_test(
            ["a", 1, "p,q"],
            "a,1,|p,q|",
            dialect=SimpleDialect(delimiter=",", quotechar="|", escapechar=""),
        )

    def test_write_csv_dialect(self) -> None:
        self._write_test(
            ["a", 1, "p,q"],
            'a,1,"p,q"',
            dialect="excel",
        )
        self._write_test(
            ["a", 1, "p,q"],
            '"a","1","p,q"',
            dialect=csv.unix_dialect,
        )
        self._write_test(
            [1, 2, 3],
            "1\t2\t3",
            dialect=clevercsv.excel_tab,
        )


if __name__ == "__main__":
    unittest.main()



import abc
import argparse

from typing import Dict
from typing import List
from typing import Optional
from typing import TextIO

class DocumentableMixin(metaclass=abc.ABCMeta):
    def __init__(
        self,
        description: Optional[str] = None,
        extra_sections: Optional[Dict[str, str]] = None,
        options_prolog: Optional[str] = None,
        options_epilog: Optional[str] = None,
    ) -> None: ...
    @property
    def description(self) -> Optional[str]: ...
    @property
    def parser(self) -> argparse.ArgumentParser: ...
    @parser.setter
    def parser(self, parser: argparse.ArgumentParser) -> None: ...
    @property
    def args(self) -> argparse.Namespace: ...
    @args.setter
    def args(self, args: argparse.Namespace) -> None: ...
    @property
    def argument_help(self) -> Dict[str, Optional[str]]: ...

class Application(DocumentableMixin):
    def __init__(
        self,
        name: str,
        version: str,
        author: Optional[str] = None,
        title: Optional[str] = None,
        description: Optional[str] = None,
        default_command: Optional[str] = None,
        add_help: bool = True,
        extra_sections: Optional[Dict[str, str]] = None,
        prolog: Optional[str] = None,
        epilog: Optional[str] = None,
        options_prolog: Optional[str] = None,
        options_epilog: Optional[str] = None,
        add_commands_section: bool = False,
    ) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def author(self) -> str: ...
    @property
    def version(self) -> str: ...
    @property
    def commands(self) -> List[Command]: ...
    @property
    def groups(self) -> List[Group]: ...
    def add_argument(self, *args, **kwargs) -> argparse.Action: ...
    def add(self, command: Command): ...
    def add_group(self, title: str) -> Group: ...
    def register(self): ...
    def handle(self) -> int: ...
    def run(
        self,
        args: Optional[List[str]] = None,
        namespace: Optional[argparse.Namespace] = None,
        exit_on_error: bool = True,
    ) -> int: ...
    def run_command(self, command: Command) -> int: ...
    def get_command(self, command_name: str) -> Command: ...
    def set_prolog(self, prolog: str) -> None: ...
    def set_epilog(self, epilog: str) -> None: ...
    def get_commands_text(self) -> str: ...
    def create_manpage(self) -> ManPage: ...
    def format_help(self) -> str: ...
    def print_help(self, file: Optional[TextIO] = None) -> None: ...

class Group:
    def __init__(
        self, title: Optional[str] = None, is_root: bool = False
    ) -> None: ...
    @property
    def application(self) -> Optional[Application]: ...
    @property
    def title(self) -> Optional[str]: ...
    @property
    def commands(self) -> List[Command]: ...
    @property
    def is_root(self) -> bool: ...
    def commands_as_actions(self) -> List[argparse.Action]: ...
    def set_app(self, app: Application) -> None: ...
    def add(self, command: Command) -> None: ...
    def __len__(self) -> int: ...

class Command(DocumentableMixin, metaclass=abc.ABCMeta):
    def __init__(
        self,
        name: str,
        title: Optional[str] = None,
        description: Optional[str] = None,
        add_help: bool = True,
        extra_sections: Optional[Dict[str, str]] = None,
        options_prolog: Optional[str] = None,
        options_epilog: Optional[str] = None,
    ) -> None: ...
    @property
    def application(self) -> Optional[Application]: ...
    @property
    def name(self) -> str: ...
    @property
    def title(self) -> Optional[str]: ...
    def add_argument(self, *args, **kwargs) -> None: ...
    def add_argument_group(self, *args, **kwargs) -> ArgumentGroup: ...
    def add_mutually_exclusive_group(
        self, *args, **kwargs
    ) -> MutuallyExclusiveGroup: ...
    def register(self) -> None: ...
    @abc.abstractmethod
    def handle(self) -> int: ...
    def create_manpage(self) -> ManPage: ...

class ManPage:
    def __init__(
        self,
        application_name: str,
        author: Optional[str] = "",
        command_name: Optional[str] = None,
        date: Optional[str] = None,
        title: Optional[str] = None,
        version: Optional[str] = "",
    ) -> None: ...
    @property
    def name(self) -> str: ...
    def metadata(self) -> List[str]: ...
    def preamble(self) -> List[str]: ...
    def header(self) -> str: ...
    def section_name(self) -> str: ...
    def add_section_synopsis(self, synopsis: str) -> None: ...
    def add_section(self, label: str, text: str) -> None: ...
    def groffify(self, text: str) -> str: ...
    def groffify_line(self, line: str) -> str: ...
    def export(self, output_dir: str) -> str: ...

class ArgumentGroup:
    def __init__(self, group: argparse._ArgumentGroup) -> None: ...
    @property
    def command(self) -> Optional[Command]: ...
    @command.setter
    def command(self, command: Command) -> None: ...
    def add_argument(self, *args, **kwargs) -> None: ...

class MutuallyExclusiveGroup:
    def __init__(self, meg: argparse._MutuallyExclusiveGroup) -> None: ...
    @property
    def command(self) -> Optional[Command]: ...
    @command.setter
    def command(self, command: Command) -> None: ...
    def add_argument(self, *args, **kwargs) -> None: ...

class Tester:
    def __init__(self, app: Application) -> None: ...
    @property
    def application(self) -> Application: ...
    def clear(self) -> None: ...
    def get_return_code(self) -> Optional[int]: ...
    def get_stdout(self) -> Optional[str]: ...
    def get_stderr(self) -> Optional[str]: ...
    def test_command(self, cmd_name: str, args: List[str]) -> None: ...
    def test_application(self, args: Optional[List[str]] = None) -> None: ...



# -*- coding: utf-8 -*-

"""
Unit tests for the pattern score.

Author: Gertjan van den Burg

"""

import unittest

from clevercsv import detect_pattern
from clevercsv.dialect import SimpleDialect


class PatternTestCase(unittest.TestCase):

    """
    Abstraction tests
    """

    def test_abstraction_1(self) -> None:
        out = detect_pattern.make_abstraction(
            "A,B,C", SimpleDialect(delimiter=",", quotechar="", escapechar="")
        )
        exp = "CDCDC"
        self.assertEqual(exp, out)

    def test_abstraction_2(self) -> None:
        out = detect_pattern.make_abstraction(
            "A,\rA,A,A\r",
            SimpleDialect(delimiter=",", quotechar="", escapechar=""),
        )
        exp = "CDCRCDCDC"
        self.assertEqual(exp, out)

    def test_abstraction_3(self) -> None:
        out = detect_pattern.make_abstraction(
            "a,a,\n,a,a\ra,a,a\r\n",
            SimpleDialect(delimiter=",", quotechar="", escapechar=""),
        )
        exp = "CDCDCRCDCDCRCDCDC"
        self.assertEqual(exp, out)

    def test_abstraction_4(self) -> None:
        out = detect_pattern.make_abstraction(
            'a,"bc""d""e""f""a",\r\n',
            SimpleDialect(delimiter=",", quotechar='"', escapechar=""),
        )
        exp = "CDCDC"
        self.assertEqual(exp, out)

    def test_abstraction_5(self) -> None:
        out = detect_pattern.make_abstraction(
            'a,"bc""d"",|"f|""',
            SimpleDialect(delimiter=",", quotechar='"', escapechar="|"),
        )
        exp = "CDC"
        self.assertEqual(exp, out)

    def test_abstraction_6(self) -> None:
        out = detect_pattern.make_abstraction(
            ",,,", SimpleDialect(delimiter=",", quotechar="", escapechar="")
        )
        exp = "CDCDCDC"
        self.assertEqual(exp, out)

    def test_abstraction_7(self) -> None:
        out = detect_pattern.make_abstraction(
            ',"",,', SimpleDialect(delimiter=",", quotechar='"', escapechar="")
        )
        exp = "CDCDCDC"
        self.assertEqual(exp, out)

    def test_abstraction_8(self) -> None:
        out = detect_pattern.make_abstraction(
            ',"",,\r\n',
            SimpleDialect(delimiter=",", quotechar='"', escapechar=""),
        )
        exp = "CDCDCDC"
        self.assertEqual(exp, out)

    """
    Escape char tests
    """

    def test_abstraction_9(self) -> None:
        out = detect_pattern.make_abstraction(
            "A,B|,C",
            SimpleDialect(delimiter=",", quotechar="", escapechar="|"),
        )
        exp = "CDC"
        self.assertEqual(exp, out)

    def test_abstraction_10(self) -> None:
        out = detect_pattern.make_abstraction(
            'A,"B,C|"D"',
            SimpleDialect(delimiter=",", quotechar='"', escapechar="|"),
        )
        exp = "CDC"
        self.assertEqual(exp, out)

    def test_abstraction_11(self) -> None:
        out = detect_pattern.make_abstraction(
            "a,|b,c",
            SimpleDialect(delimiter=",", quotechar="", escapechar="|"),
        )
        exp = "CDCDC"
        self.assertEqual(exp, out)

    def test_abstraction_12(self) -> None:
        out = detect_pattern.make_abstraction(
            "a,b|,c",
            SimpleDialect(delimiter=",", quotechar="", escapechar="|"),
        )
        exp = "CDC"
        self.assertEqual(exp, out)

    def test_abstraction_13(self) -> None:
        out = detect_pattern.make_abstraction(
            'a,"b,c|""',
            SimpleDialect(delimiter=",", quotechar='"', escapechar="|"),
        )
        exp = "CDC"
        self.assertEqual(exp, out)

    def test_abstraction_14(self) -> None:
        out = detect_pattern.make_abstraction(
            "a,b||c",
            SimpleDialect(delimiter=",", quotechar="", escapechar="|"),
        )
        exp = "CDC"
        self.assertEqual(exp, out)

    def test_abstraction_15(self) -> None:
        out = detect_pattern.make_abstraction(
            'a,"b|"c||d|"e"',
            SimpleDialect(delimiter=",", quotechar='"', escapechar="|"),
        )
        exp = "CDC"
        self.assertEqual(exp, out)

    def test_abstraction_16(self) -> None:
        out = detect_pattern.make_abstraction(
            'a,"b|"c||d","e"',
            SimpleDialect(delimiter=",", quotechar='"', escapechar="|"),
        )
        exp = "CDCDC"
        self.assertEqual(exp, out)

    """
    Fill empties
    """

    def test_fill_empties_1(self) -> None:
        out = detect_pattern.fill_empties("DDD")
        exp = "CDCDCDC"
        self.assertEqual(exp, out)

    """
    Pattern Score tests
    """

    def test_pattern_score_1(self) -> None:
        # theta_1 from paper
        data = (
            "7,5; Mon, Jan 12;6,40\n100; Fri, Mar 21;8,23\n8,2; Thu, Sep 17;"
            '2,71\n538,0;;7,26\n"NA"; Wed, Oct 4;6,93'
        )
        d = SimpleDialect(delimiter=",", quotechar="", escapechar="")
        out = detect_pattern.pattern_score(data, d)
        exp = 7 / 4
        self.assertAlmostEqual(exp, out)

    def test_pattern_score_2(self) -> None:
        # theta_2 from paper
        data = (
            "7,5; Mon, Jan 12;6,40\n100; Fri, Mar 21;8,23\n8,2; Thu, Sep 17;"
            '2,71\n538,0;;7,26\n"NA"; Wed, Oct 4;6,93'
        )
        d = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        out = detect_pattern.pattern_score(data, d)
        exp = 10 / 3
        self.assertAlmostEqual(exp, out)

    def test_pattern_score_3(self) -> None:
        # theta_3 from paper
        data = (
            "7,5; Mon, Jan 12;6,40\n100; Fri, Mar 21;8,23\n8,2; Thu, Sep 17;"
            '2,71\n538,0;;7,26\n"NA"; Wed, Oct 4;6,93'
        )
        d = SimpleDialect(delimiter=";", quotechar='"', escapechar="")
        out = detect_pattern.pattern_score(data, d)
        exp = 10 / 3
        self.assertAlmostEqual(exp, out)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

"""
Script to run PythonFuzz to detect unhandled exceptions in the Sniffer

This file is part of CleverCSV.

"""

from pythonfuzz.main import PythonFuzz

import clevercsv


@PythonFuzz
def fuzz(buf):
    try:
        string = buf.decode("utf-8")
        _ = clevercsv.Sniffer().sniff(string)
    except UnicodeDecodeError:
        pass
    except clevercsv.exceptions.Error:
        pass


if __name__ == "__main__":
    fuzz()



# -*- coding: utf-8 -*-

"""
Unit test for consistency score

Author: G.J.J. van den Burg

"""

import unittest

from clevercsv.consistency import ConsistencyDetector
from clevercsv.consistency import ConsistencyScore
from clevercsv.dialect import SimpleDialect


class ConsistencyTestCase(unittest.TestCase):
    def test_get_best_set_1(self) -> None:
        scores = {
            SimpleDialect(",", None, None): ConsistencyScore(P=1, T=1, Q=1),
            SimpleDialect(";", None, None): ConsistencyScore(
                P=1, T=None, Q=None
            ),
            SimpleDialect("|", None, None): ConsistencyScore(P=2, T=1, Q=2),
        }
        H = ConsistencyDetector.get_best_dialects(scores)
        self.assertEqual(H, [SimpleDialect("|", None, None)])



# -*- coding: utf-8 -*-

"""
Drop-in replacement for Python Sniffer object.

Author: Gertjan van den Burg

"""

from enum import Enum
from io import StringIO

from typing import Dict
from typing import Iterable
from typing import Optional
from typing import Union

from .consistency import ConsistencyDetector
from .dialect import SimpleDialect
from .exceptions import NoDetectionResult
from .normal_form import detect_dialect_normal
from .read import reader


class DetectionMethod(str, Enum):
    """Possible detection methods

    Valid options are `"auto"` (the default for :class:`Detector.detect`),
    `"normal"`, or `"consistency"`.  The `"auto"` option first attempts to
    detect the dialect using normal-form detection, and uses the consistency
    measure if normal-form detection is inconclusive. The `"normal"` method
    uses normal-form detection excllusively, and the `"consistency"` method
    uses the consistency measure exclusively.
    """

    AUTO = "auto"
    NORMAL = "normal"
    CONSISTENCY = "consistency"


class Detector:
    """
    Detect the Dialect of CSV files with normal forms or the data consistency
    measure. This class provides a drop-in replacement for the Python dialect
    Sniffer from the standard library.

    Note
    ----
    We call the object ``Detector`` just to mark the difference in the
    implementation and avoid naming issues. You can import it as ``from ccsv
    import Sniffer`` nonetheless.

    """

    def sniff(
        self,
        sample: str,
        delimiters: Optional[Iterable[str]] = None,
        verbose: bool = False,
    ) -> Optional[SimpleDialect]:
        # Compatibility method for Python
        return self.detect(sample, delimiters=delimiters, verbose=verbose)

    def detect(
        self,
        sample: str,
        delimiters: Optional[Iterable[str]] = None,
        verbose: bool = False,
        method: Union[DetectionMethod, str] = DetectionMethod.AUTO,
        skip: bool = True,
    ) -> Optional[SimpleDialect]:
        """Detect the dialect of a CSV file

        This method detects the dialect of the CSV file using the specified
        detection method.

        Parameters
        ----------
        sample : str
            A sample of text from the CSV file. For best results and if time
            allows, use the entire contents of the CSV file as the sample.

        delimiters : Optional[Iterable[str]]
            Set of delimiters to consider for dialect detection. The potential
            dialects will be constructed by analyzing the sample and these
            delimiters. If omitted, the set of potential delimiters will be
            constructed from the sample.

        verbose : bool
            Enable verbose mode.

        method : Union[DetectionMethod, str]
            The method to use for dialect detection. Possible values are
            :class:`DetectionMethod` instances or strings that can be cast to
            as such an enum.

        skip : bool
            Whether to skip potential dialects that have too low a pattern
            score in the consistency detection. See
            :func:`ConsistencyDetector.compute_consistency_scores` for more
            details.

        Returns
        -------
        dialect : Optional[SimpleDialect]
            The detected dialect. Can be `None` if dialect detection was
            inconclusive.

        """
        method = DetectionMethod(method) if isinstance(method, str) else method
        if delimiters is not None:
            delimiters = list(delimiters)
        if method == DetectionMethod.NORMAL or method == DetectionMethod.AUTO:
            if verbose:
                print("Running normal form detection ...", flush=True)
            dialect = detect_dialect_normal(
                sample, delimiters=delimiters, verbose=verbose
            )
            if dialect is not None:
                self.method_ = DetectionMethod.NORMAL
                return dialect

        self.method_ = DetectionMethod.CONSISTENCY
        consistency_detector = ConsistencyDetector(skip=skip, verbose=verbose)
        if verbose:
            print("Running data consistency measure ...", flush=True)
        return consistency_detector.detect(sample, delimiters=delimiters)

    def has_header(self, sample: str, max_rows_to_check: int = 20) -> bool:
        """Detect if a file has a header from a sample.

        This function is copied from CPython! The only change we've made is to
        use our dialect detection method.

        """

        # Creates a dictionary of types of data in each column. If any
        # column is of a single type (say, integers), *except* for the first
        # row, then the first row is presumed to be labels. If the type
        # can't be determined, it is assumed to be a string in which case
        # the length of the string is the determining factor: if all of the
        # rows except for the first are the same length, it's a header.
        # Finally, a 'vote' is taken at the end for each column, adding or
        # subtracting from the likelihood of the first row being a header.

        dialect = self.sniff(sample)
        if dialect is None:
            raise NoDetectionResult

        rdr = reader(StringIO(sample), dialect)

        header = next(rdr)  # assume first row is header

        columns = len(header)
        columnTypes: Dict[int, Optional[Union[int, type]]] = {}
        for i in range(columns):
            columnTypes[i] = None

        thisType: Union[int, type]
        checked = 0
        for row in rdr:
            # arbitrary number of rows to check, to keep it sane
            if checked > max_rows_to_check:
                break
            checked += 1

            if len(row) != columns:
                continue  # skip rows that have irregular number of columns

            for col in list(columnTypes.keys()):
                for thisType in [int, float, complex]:
                    try:
                        thisType(row[col])
                        break
                    except (ValueError, OverflowError):
                        pass
                else:
                    # fallback to length of string
                    thisType = len(row[col])

                if thisType != columnTypes[col]:
                    if columnTypes[col] is None:  # add new column type
                        columnTypes[col] = thisType
                    else:
                        # type is inconsistent, remove column from
                        # consideration
                        del columnTypes[col]

        # finally, compare results against first row and "vote"
        # on whether it's a header
        hasHeader = 0
        for col, colType in columnTypes.items():
            if isinstance(colType, int):  # it's a length
                if len(header[col]) != colType:
                    hasHeader += 1
                else:
                    hasHeader -= 1
            else:  # attempt typecast
                if colType is None:
                    hasHeader += 1
                    continue

                try:
                    colType(header[col])
                except (ValueError, TypeError):
                    hasHeader += 1
                else:
                    hasHeader -= 1

        return hasHeader > 0



# -*- coding: utf-8 -*-

"""
Unit tests for the console application.

Author: Gertjan van den Burg
"""

import json
import os
import tempfile
import unittest

from typing import Any
from typing import List
from typing import Optional

from wilderness import Tester

from clevercsv import __version__
from clevercsv._types import _DialectLike
from clevercsv.console import build_application
from clevercsv.dialect import SimpleDialect
from clevercsv.encoding import get_encoding
from clevercsv.write import writer

TableType = List[List[Any]]


class ConsoleTestCase(unittest.TestCase):
    def _build_file(
        self,
        table: TableType,
        dialect: _DialectLike,
        encoding: Optional[str] = None,
        newline: Optional[str] = None,
    ) -> str:
        tmpfd, tmpfname = tempfile.mkstemp(
            prefix="ccsv_",
            suffix=".csv",
        )
        tmpid = os.fdopen(tmpfd, "w", newline=newline, encoding=encoding)
        w = writer(tmpid, dialect=dialect)
        w.writerows(table)
        tmpid.close()
        return tmpfname

    def _detect_test_wrap(
        self, table: TableType, dialect: _DialectLike
    ) -> None:
        tmpfname = self._build_file(table, dialect)
        exp = "Detected: " + str(dialect)

        application = build_application()
        tester = Tester(application)
        tester.test_command("detect", [tmpfname])

        try:
            stdout = tester.get_stdout()
            self.assertIsNotNone(stdout)
            assert stdout is not None
            output = stdout.strip()
            self.assertEqual(exp, output)
        finally:
            os.unlink(tmpfname)

    def test_detect_base(self) -> None:
        table: TableType = [["A", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        with self.subTest(name="simple"):
            self._detect_test_wrap(table, dialect)

        table = [["A,0", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=",", quotechar="", escapechar="\\")
        with self.subTest(name="escaped"):
            self._detect_test_wrap(table, dialect)

        table = [["A,0", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=",", quotechar='"', escapechar="")
        with self.subTest(name="quoted"):
            self._detect_test_wrap(table, dialect)

        table = [['a"A,0"b', "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=",", quotechar='"', escapechar="")
        with self.subTest(name="double"):
            self._detect_test_wrap(table, dialect)

    def test_detect_opts_1(self) -> None:
        table: TableType = [["A", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        encoding = "windows-1252"
        tmpfname = self._build_file(table, dialect, encoding=encoding)

        application = build_application()
        tester = Tester(application)
        tester.test_command("detect", ["--encoding", encoding, tmpfname])

        exp = "Detected: " + str(dialect)

        try:
            stdout = tester.get_stdout()
            self.assertIsNotNone(stdout)
            assert stdout is not None
            output = stdout.strip()
            self.assertEqual(exp, output)
        finally:
            os.unlink(tmpfname)

    def test_detect_opts_2(self) -> None:
        table: TableType = [["A", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        tmpfname = self._build_file(table, dialect)

        application = build_application()
        tester = Tester(application)
        tester.test_command("detect", ["--num-chars", "5", tmpfname])

        exp = "Detected: " + str(dialect)

        try:
            stdout = tester.get_stdout()
            self.assertIsNotNone(stdout)
            assert stdout is not None
            output = stdout.strip()
            self.assertEqual(exp, output)
        finally:
            os.unlink(tmpfname)

    def test_detect_opts_3(self) -> None:
        table: TableType = [["A", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        tmpfname = self._build_file(table, dialect)

        application = build_application()
        tester = Tester(application)
        tester.test_command("detect", ["--plain", tmpfname])

        exp = """\
delimiter = ;
quotechar =
escapechar ="""
        try:
            stdout = tester.get_stdout()
            self.assertIsNotNone(stdout)
            assert stdout is not None
            output = stdout.strip()
            self.assertEqual(exp, output)
        finally:
            os.unlink(tmpfname)

    def test_detect_opts_4(self) -> None:
        table: TableType = [["A", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        tmpfname = self._build_file(table, dialect)

        application = build_application()
        tester = Tester(application)
        tester.test_command("detect", ["--json", "--add-runtime", tmpfname])

        try:
            stdout = tester.get_stdout()
            self.assertIsNotNone(stdout)
            assert stdout is not None
            output = stdout.strip()
            data = json.loads(output)
            self.assertEqual(data["delimiter"], ";")
            self.assertEqual(data["quotechar"], "")
            self.assertEqual(data["escapechar"], "")
            self.assertIsInstance(data["runtime"], float)
        finally:
            os.unlink(tmpfname)

    def test_code_1(self) -> None:
        table: TableType = [["A", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")

        tmpfname = self._build_file(table, dialect)

        application = build_application()
        tester = Tester(application)
        tester.test_command("code", [tmpfname])

        # for chardet
        exp_1 = f"""\

# Code generated with CleverCSV version {__version__}

import clevercsv

with open("{tmpfname}", "r", newline="", encoding="ascii") as fp:
    reader = clevercsv.reader(fp, delimiter=";", quotechar="", escapechar="")
    rows = list(reader)

"""

        # for cChardet
        exp_2 = f"""\

# Code generated with CleverCSV version {__version__}

import clevercsv

with open("{tmpfname}", "r", newline="", encoding="ASCII") as fp:
    reader = clevercsv.reader(fp, delimiter=";", quotechar="", escapechar="")
    rows = list(reader)

"""
        try:
            output = tester.get_stdout()
            self.assertIn(output, [exp_1, exp_2])
        finally:
            os.unlink(tmpfname)

    def test_code_2(self) -> None:
        table: TableType = [["A", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        tmpfname = self._build_file(table, dialect)

        application = build_application()
        tester = Tester(application)
        tester.test_command("code", ["-p", tmpfname])

        exp = f"""\

# Code generated with CleverCSV version {__version__}

import clevercsv

df = clevercsv.read_dataframe("{tmpfname}", delimiter=";", quotechar="", escapechar="")

"""
        try:
            output = tester.get_stdout()
            self.assertEqual(exp, output)
        finally:
            os.unlink(tmpfname)

    def test_code_3(self) -> None:
        table: TableType = [["Å", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        encoding = "ISO-8859-1"
        tmpfname = self._build_file(table, dialect, encoding=encoding)

        application = build_application()
        tester = Tester(application)
        tester.test_command("code", [tmpfname])

        # for chardet
        exp_1 = f"""\

# Code generated with CleverCSV version {__version__}

import clevercsv

with open("{tmpfname}", "r", newline="", encoding="{encoding}") as fp:
    reader = clevercsv.reader(fp, delimiter=";", quotechar="", escapechar="")
    rows = list(reader)

"""

        # for cChardet
        exp_2 = f"""\

# Code generated with CleverCSV version {__version__}

import clevercsv

with open("{tmpfname}", "r", newline="", encoding="WINDOWS-1252") as fp:
    reader = clevercsv.reader(fp, delimiter=";", quotechar="", escapechar="")
    rows = list(reader)

"""
        try:
            output = tester.get_stdout()
            self.assertIn(output, [exp_1, exp_2])
        finally:
            os.unlink(tmpfname)

    def test_code_4(self) -> None:
        table: TableType = [["Å", "B,D", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=",", quotechar="", escapechar="\\")
        encoding = "ISO-8859-1"
        tmpfname = self._build_file(table, dialect, encoding=encoding)

        application = build_application()
        tester = Tester(application)
        tester.test_command("code", [tmpfname])

        # for chardet
        exp_1 = f"""\

# Code generated with CleverCSV version {__version__}

import clevercsv

with open("{tmpfname}", "r", newline="", encoding="{encoding}") as fp:
    reader = clevercsv.reader(fp, delimiter=",", quotechar="", escapechar="\\\\")
    rows = list(reader)

"""

        # for cChardet
        exp_2 = f"""\

# Code generated with CleverCSV version {__version__}

import clevercsv

with open("{tmpfname}", "r", newline="", encoding="WINDOWS-1252") as fp:
    reader = clevercsv.reader(fp, delimiter=",", quotechar="", escapechar="\\\\")
    rows = list(reader)

"""
        try:
            output = tester.get_stdout()
            self.assertIn(output, [exp_1, exp_2])
        finally:
            os.unlink(tmpfname)

    def test_code_5(self) -> None:
        table: TableType = [["A", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter="\t", quotechar="", escapechar="")

        tmpfname = self._build_file(table, dialect)

        application = build_application()
        tester = Tester(application)
        tester.test_command("code", [tmpfname])

        # for chardet
        exp_1 = f"""\

# Code generated with CleverCSV version {__version__}

import clevercsv

with open("{tmpfname}", "r", newline="", encoding="ascii") as fp:
    reader = clevercsv.reader(fp, delimiter="\\t", quotechar="", escapechar="")
    rows = list(reader)

"""

        # for cChardet
        exp_2 = f"""\

# Code generated with CleverCSV version {__version__}

import clevercsv

with open("{tmpfname}", "r", newline="", encoding="ASCII") as fp:
    reader = clevercsv.reader(fp, delimiter="\\t", quotechar="", escapechar="")
    rows = list(reader)

"""

        try:
            output = tester.get_stdout()
            self.assertIn(output, [exp_1, exp_2])
        finally:
            os.unlink(tmpfname)

    def test_standardize_1(self) -> None:
        table: TableType = [["A", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        tmpfname = self._build_file(table, dialect)

        application = build_application()
        tester = Tester(application)
        tester.test_command("standardize", [tmpfname])

        # Excel format (i.e. RFC4180) *requires* CRLF
        crlf = "\r\n"
        exp = crlf.join(["A,B,C", "1,2,3", "4,5,6"])
        # add line terminator of last row
        exp += crlf
        try:
            output = tester.get_stdout()
            self.assertEqual(exp, output)
        finally:
            os.unlink(tmpfname)

    def test_standardize_2(self) -> None:
        table: TableType = [["A", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        tmpfname = self._build_file(table, dialect)

        tmpfd, tmpoutname = tempfile.mkstemp(prefix="ccsv_", suffix=".csv")
        os.close(tmpfd)

        application = build_application()
        tester = Tester(application)
        tester.test_command("standardize", ["-o", tmpoutname, tmpfname])

        # Excel format (i.e. RFC4180) *requires* CRLF
        crlf = "\r\n"
        exp = crlf.join(["A,B,C", "1,2,3", "4,5,6", ""])
        with open(tmpoutname, "r", newline="") as fp:
            output = fp.read()

        try:
            self.assertEqual(exp, output)
        finally:
            os.unlink(tmpfname)
            os.unlink(tmpoutname)

    def test_standardize_3(self) -> None:
        table: TableType = [["A", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        tmpfname = self._build_file(table, dialect)

        application = build_application()
        tester = Tester(application)
        tester.test_command("standardize", ["-t", tmpfname])

        # Excel format (i.e. RFC4180) *requires* CRLF
        crlf = "\r\n"
        exp = crlf.join(["A,1,4", "B,2,5", "C,3,6"])
        # add line terminator of last row
        exp += crlf

        try:
            output = tester.get_stdout()
            self.assertEqual(exp, output)
        finally:
            os.unlink(tmpfname)

    def test_standardize_in_place(self) -> None:
        table: TableType = [["A", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        tmpfname = self._build_file(table, dialect)

        application = build_application()
        tester = Tester(application)
        tester.test_command("standardize", ["-i", tmpfname])

        self.assertEqual(tester.get_return_code(), 2)

        # Excel format (i.e. RFC4180) *requires* CRLF
        crlf = "\r\n"
        exp = crlf.join(["A,B,C", "1,2,3", "4,5,6"])
        # add line terminator of last row
        exp += crlf

        try:
            with open(tmpfname, "r", newline="") as fp:
                contents = fp.read()
            self.assertEqual(exp, contents)
        finally:
            os.unlink(tmpfname)

    def test_standardize_in_place_noop(self) -> None:
        table: TableType = [["A", "B", "C"], [1, 2, 3], [4, 5, 6]]
        dialect = "excel"
        tmpfname = self._build_file(table, dialect, newline="")

        application = build_application()
        tester = Tester(application)
        tester.test_command("standardize", ["-i", tmpfname])

        self.assertEqual(tester.get_return_code(), 0)

        # Excel format (i.e. RFC4180) *requires* CRLF
        crlf = "\r\n"
        exp = crlf.join(["A,B,C", "1,2,3", "4,5,6"])
        # add line terminator of last row
        exp += crlf

        try:
            with open(tmpfname, "r", newline="") as fp:
                contents = fp.read()
            self.assertEqual(exp, contents)
        finally:
            os.unlink(tmpfname)

    def test_standardize_multi(self) -> None:
        table: TableType = [
            ["A", "B", "C"],
            [1, 2, 3],
            [4, 5, 6],
        ]
        dialects = ["excel", "unix", "excel-tab"]
        tmpfnames = [self._build_file(table, D, newline="") for D in dialects]

        tmpoutnames = []
        for _ in range(len(dialects)):
            tmpfd, tmpoutname = tempfile.mkstemp(prefix="ccsv_", suffix=".csv")
            os.close(tmpfd)
            tmpoutnames.append(tmpoutname)

        args = []
        for tmpoutname in tmpoutnames:
            args.append("-o")
            args.append(tmpoutname)
        args += tmpfnames

        application = build_application()
        tester = Tester(application)
        tester.test_command("standardize", args)

        self.assertEqual(tester.get_return_code(), 0)

        exp = "\r\n".join([",".join(map(str, row)) for row in table]) + "\r\n"

        try:
            for tmpoutname in tmpoutnames:
                with open(tmpoutname, "r", newline="") as fp:
                    contents = fp.read()
                self.assertEqual(contents, exp)
        finally:
            any(map(os.unlink, tmpfnames))
            any(map(os.unlink, tmpoutnames))

    def test_standardize_multi_errors(self) -> None:
        table: TableType = [
            ["A", "B", "C"],
            [1, 2, 3],
            [4, 5, 6],
        ]
        dialects = ["excel", "unix", "excel-tab"]
        tmpfnames = [self._build_file(table, D, newline="") for D in dialects]

        tmpoutnames = []
        for _ in range(1):
            tmpfd, tmpoutname = tempfile.mkstemp(prefix="ccsv_", suffix=".csv")
            os.close(tmpfd)
            tmpoutnames.append(tmpoutname)

        args = []
        for tmpoutname in tmpoutnames:
            args.append("-o")
            args.append(tmpoutname)
        args += tmpfnames

        application = build_application()
        tester = Tester(application)
        tester.test_command("standardize", args)

        self.assertEqual(tester.get_return_code(), 1)

        self.assertEqual(
            tester.get_stderr(),
            "Number of output files should match the number of input files.\n",
        )

        any(map(os.unlink, tmpfnames))
        any(map(os.unlink, tmpoutnames))

    def test_standardize_multi_encoding(self) -> None:
        table: TableType = [
            ["Å", "B", "C"],
            [1, 2, 3],
            [4, 5, 6],
        ]
        dialects = ["excel", "unix", "excel-tab"]
        encoding = "ISO-8859-1"
        tmpfnames = [
            self._build_file(table, D, newline="", encoding=encoding)
            for D in dialects
        ]

        tmpoutnames = []
        for _ in range(len(dialects)):
            tmpfd, tmpoutname = tempfile.mkstemp(prefix="ccsv_", suffix=".csv")
            os.close(tmpfd)
            tmpoutnames.append(tmpoutname)

        args = ["-e", encoding]
        for tmpoutname in tmpoutnames:
            args.append("-o")
            args.append(tmpoutname)
        args += tmpfnames

        application = build_application()
        tester = Tester(application)
        tester.test_command("standardize", args)

        self.assertEqual(tester.get_return_code(), 0)

        exp = "\r\n".join([",".join(map(str, row)) for row in table]) + "\r\n"

        try:
            for tmpoutname in tmpoutnames:
                with open(
                    tmpoutname, "r", newline="", encoding=encoding
                ) as fp:
                    contents = fp.read()
                self.assertEqual(contents, exp)
        finally:
            any(map(os.unlink, tmpfnames))
            any(map(os.unlink, tmpoutnames))

    def test_standardize_in_place_multi(self) -> None:
        table: TableType = [
            ["Å", "B", "C"],
            [1, 2, 3],
            [4, 5, 6],
        ]
        dialects = ["excel", "unix", "excel-tab"]
        encoding = "ISO-8859-1"
        tmpfnames = [
            self._build_file(table, D, newline="", encoding=encoding)
            for D in dialects
        ]

        application = build_application()
        tester = Tester(application)
        tester.test_command("standardize", ["-i", "-e", encoding, *tmpfnames])

        self.assertEqual(tester.get_return_code(), 2)

        exp = "\r\n".join([",".join(map(str, row)) for row in table]) + "\r\n"

        try:
            for tmpfname in tmpfnames:
                with open(tmpfname, "r", newline="", encoding=encoding) as fp:
                    contents = fp.read()
                self.assertEqual(contents, exp)
        finally:
            any(map(os.unlink, tmpfnames))

    def test_standardize_in_place_multi_noop(self) -> None:
        table: TableType = [
            ["Å", "B", "C"],
            [1, 2, 3],
            [4, 5, 6],
        ]
        dialects = ["excel", "excel", "excel"]
        tmpfnames = [self._build_file(table, D, newline="") for D in dialects]

        application = build_application()
        tester = Tester(application)
        tester.test_command("standardize", ["-i", *tmpfnames])

        self.assertEqual(tester.get_return_code(), 0)

        exp = "\r\n".join([",".join(map(str, row)) for row in table]) + "\r\n"
        try:
            for tmpfname in tmpfnames:
                with open(tmpfname, "r", newline="") as fp:
                    contents = fp.read()
                self.assertEqual(contents, exp)
        finally:
            any(map(os.unlink, tmpfnames))

    def test_standardize_target_encoding(self) -> None:
        table: TableType = [["Å", "B", "C"], ["é", "ü", "中"], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        encoding = "utf-8"
        tmpfname = self._build_file(table, dialect, encoding=encoding)

        tmpfd, tmpoutname = tempfile.mkstemp(prefix="ccsv_", suffix=".csv")
        os.close(tmpfd)

        application = build_application()
        tester = Tester(application)
        tester.test_command(
            "standardize", ["-o", tmpoutname, "-E", "utf-8", tmpfname]
        )

        # Excel format (i.e. RFC4180) *requires* CRLF
        crlf = "\r\n"
        exp = crlf.join(["Å,B,C", "é,ü,中", "4,5,6", ""])
        with open(tmpoutname, "r", newline="", encoding="utf-8") as fp:
            output = fp.read()

        try:
            self.assertEqual(exp, output)
        finally:
            os.unlink(tmpfname)
            os.unlink(tmpoutname)

    def test_standardize_target_encoding2(self) -> None:
        table: TableType = [["A", "B", "C"], ["é", "è", "à"], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        encoding = "latin-1"
        tmpfname = self._build_file(table, dialect, encoding=encoding)
        self.assertEqual(
            "ISO-8859-1", get_encoding(tmpfname, try_cchardet=False)
        )
        tmpfd, tmpoutname = tempfile.mkstemp(prefix="ccsv_", suffix=".csv")
        os.close(tmpfd)

        application = build_application()
        tester = Tester(application)
        tester.test_command(
            "standardize",
            ["-o", tmpoutname, "-e", "latin-1", "-E", "utf-8", tmpfname],
        )

        # Excel format (i.e. RFC4180) *requires* CRLF
        crlf = "\r\n"
        exp = crlf.join(["A,B,C", "é,è,à", "4,5,6", ""])

        self.assertEqual("utf-8", get_encoding(tmpoutname, try_cchardet=False))
        with open(tmpoutname, "r", newline="", encoding="utf-8") as fp:
            output = fp.read()

        try:
            self.assertEqual(exp, output)

        finally:
            os.unlink(tmpfname)
            os.unlink(tmpoutname)

    def test_standardize_target_encoding_raise_UnicodeEncodeError(
        self,
    ) -> None:
        table: TableType = [["Å", "B", "C"], ["é", "ü", "中"], [4, 5, 6]]
        dialect = SimpleDialect(delimiter=";", quotechar="", escapechar="")
        encoding = "utf-8"
        tmpfname = self._build_file(table, dialect, encoding=encoding)

        tmpfd, tmpoutname = tempfile.mkstemp(prefix="ccsv_", suffix=".csv")
        os.close(tmpfd)

        application = build_application()
        tester = Tester(application)
        try:
            with self.assertRaises(UnicodeEncodeError):
                tester.test_command(
                    "standardize",
                    ["-o", tmpoutname, "-E", "latin-1", tmpfname],
                )
        finally:
            os.unlink(tmpfname)
            os.unlink(tmpoutname)



# -*- coding: utf-8 -*-

import sys

from typing import List
from typing import Optional
from typing import Sequence

from wilderness import Command

from clevercsv._optional import import_optional_dependency
from clevercsv.exceptions import NoDetectionResult
from clevercsv.wrappers import read_table

from ._docs import FLAG_DESCRIPTIONS
from ._utils import parse_int


class ViewCommand(Command):
    _description = (
        "The view command is useful to quickly inspect a messy CSV file on "
        "the command line."
    )

    def __init__(self) -> None:
        super().__init__(
            name="view",
            title="View the CSV file on the command line using TabView",
            description=self._description,
            extra_sections={"CleverCSV": "Part of the CleverCSV suite"},
        )

    def register(self) -> None:
        self.add_argument("path", help="Path to the CSV file")
        self.add_argument(
            "-e",
            "--encoding",
            help="Set the encoding of the file",
            description=FLAG_DESCRIPTIONS["encoding"],
        )
        self.add_argument(
            "-n",
            "--num-chars",
            help="Number of characters to use for detection",
            type=int,
            description=FLAG_DESCRIPTIONS["num-chars"],
        )
        self.add_argument(
            "-t",
            "--transpose",
            action="store_true",
            help="Transpose the columns of the input file before viewing",
        )

    def _tabview(self, rows: List[List[str]]) -> None:
        if sys.platform == "win32":
            print(
                "Error: unfortunately Tabview is not available on Windows, so "
                "the clevercsv view command is not available",
                file=sys.stderr,
            )
            return

        import_optional_dependency("tabview", raise_on_missing=True)
        from tabview import view

        view(rows)

    def handle(self) -> int:
        verbose = self.args.verbose
        num_chars = parse_int(self.args.num_chars, "num-chars")
        try:
            rows = read_table(
                self.args.path,
                encoding=self.args.encoding,
                num_chars=num_chars,
                verbose=verbose,
            )
        except NoDetectionResult:
            print("Error: dialect detection failed.", file=sys.stderr)
            return 1

        if self.args.transpose:
            max_row_length = max(map(len, rows))
            fixed_rows: List[Sequence[Optional[str]]] = []
            for row in rows:
                if len(row) == max_row_length:
                    fixed_rows.append(row)
                else:
                    fixed_rows.append(
                        row + [None] * (max_row_length - len(row))
                    )
            rows = list(map(list, zip(*fixed_rows)))
        self._tabview(rows)
        return 0



# -*- coding: utf-8 -*-

import sys

from wilderness import Command

from clevercsv.encoding import get_encoding
from clevercsv.wrappers import detect_dialect

from ._docs import FLAG_DESCRIPTIONS
from ._utils import generate_code
from ._utils import parse_int


class CodeCommand(Command):
    _description = (
        "Generate Python code for importing a given CSV file. This is "
        "especially useful if you don't want to repeatedly detect the dialect "
        "of the same file. Simply run:\n\n"
        "\tclevercsv code your_csv_file.csv\n\n"
        "and copy the generated code to a Python script."
    )

    def __init__(self) -> None:
        super().__init__(
            name="code",
            title="Generate Python code to import a CSV file",
            description=self._description,
            extra_sections={"CleverCSV": "Part of the CleverCSV suite"},
        )

    def register(self) -> None:
        self.add_argument("path", help="Path to the CSV file")
        self.add_argument(
            "-e",
            "--encoding",
            help="Set the encoding of the file",
            description=FLAG_DESCRIPTIONS["encoding"],
        )
        self.add_argument(
            "-n",
            "--num-chars",
            type=int,
            help="Number of characters to use for detection",
            description=FLAG_DESCRIPTIONS["num-chars"],
        )
        self.add_argument(
            "-p",
            "--pandas",
            action="store_true",
            help="Write code that uses a Pandas DataFrame",
            description=(
                "By default, this command writes a small Python script to "
                "import the CSV file as a list of lists. By enabling this "
                "option the script will be written such that the file will be "
                "read as a Pandas DataFrame instead."
            ),
        )

    def handle(self) -> int:
        filename = self.args.path
        encoding = self.args.encoding or get_encoding(filename)
        num_chars = parse_int(self.args.num_chars, "num-chars")
        dialect = detect_dialect(
            filename,
            num_chars=num_chars,
            encoding=encoding,
            verbose=self.args.verbose,
        )
        if dialect is None:
            print("Error: dialect detection failed.", file=sys.stderr)
            return 1

        code_lines = generate_code(
            filename, dialect, encoding, use_pandas=self.args.pandas
        )
        print("\n".join(code_lines))
        return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Do-nothing script for making a release

This idea comes from here: 
https://blog.danslimmon.com/2019/07/15/do-nothing-scripting-the-key-to-gradual-automation/

Author: Gertjan van den Burg
Date: 2019-07-23

"""

import abc
import os
import re
import subprocess
import sys
import tempfile
import webbrowser

from pathlib import Path

from typing import Dict
from typing import Optional

import colorama

URLS = {
    "RTD": "https://readthedocs.org/projects/clevercsv/builds/",
    "CI": "https://github.com/alan-turing-institute/CleverCSV/actions",
    "dummy": "https://github.com/alan-turing-institute/CleverCSV-pre-commit",
    "tags": "https://github.com/alan-turing-institute/CleverCSV/tags",
}
BRANCH_NAME = "master"
CHANGELOG_FILENAME = "CHANGELOG.md"
MAN_DIRECTORY = "./man"
PACKAGE_NICE_NAME = "CleverCSV"


def color_text(
    msg: str,
    color: Optional[str] = None,
    style: Optional[str] = None,
) -> str:
    colors = {
        "red": colorama.Fore.RED,
        "green": colorama.Fore.GREEN,
        "cyan": colorama.Fore.CYAN,
        "yellow": colorama.Fore.YELLOW,
        "magenta": colorama.Fore.MAGENTA,
        None: "",
    }
    styles = {
        "bright": colorama.Style.BRIGHT,
        "dim": colorama.Style.DIM,
        None: "",
    }
    pre = colors[color] + styles[style]
    post = colorama.Style.RESET_ALL
    return f"{pre}{msg}{post}"


def color_print(msg, color=None, style=None):
    print(color_text(msg, color=color, style=style))


def wait_for_enter():
    input(color_text("\nPress Enter to continue", style="dim"))
    print()


def get_package_name():
    with open("./setup.py", "r") as fp:
        nameline = next(
            (line.strip() for line in fp if line.startswith("NAME = ")), None
        )
        return nameline.split("=")[-1].strip().strip('"')


def get_package_version(pkgname):
    ctx = {}
    with open(f"{pkgname.lower()}/__version__.py", "r") as fp:
        exec(fp.read(), ctx)
    return ctx["__version__"]


def get_last_version_tag():
    output = ""
    with subprocess.Popen(
        "git tag -l", shell=True, stdout=subprocess.PIPE, text=True, bufsize=1
    ) as p:
        for line in p.stdout:
            output += line
    tags = output.rstrip().split()
    version_tags = [
        t for t in tags if re.match(r"v\d+\.\d+\.\d+$", t) is not None
    ]
    versions = [v.lstrip(".") for v in version_tags]
    versions.sort()
    return versions[-1]


def get_last_release_candidate_tag(version: str):
    output = ""
    with subprocess.Popen(
        "git tag -l", shell=True, stdout=subprocess.PIPE, text=True, bufsize=1
    ) as p:
        for line in p.stdout:
            output += line
    tags = output.rstrip().split()
    version_tags = [
        t
        for t in tags
        if re.match(r"v" + version + r"-rc\.\d+", t) is not None
    ]
    versions = [v.lstrip(".") for v in version_tags]
    versions.sort()
    if not versions:
        return None
    return versions[-1]


def read_changelog_section(context: Dict[str, str]) -> str:
    next_version = context["next_version"]
    section_lines = []
    record = False
    with open(CHANGELOG_FILENAME, "r") as fileobj:
        for line in fileobj:
            if line.startswith(f"## Version {next_version}"):
                section_lines.append(line)
                record = True
                continue
            if record and line.startswith("## Version"):
                record = False
                break
            elif record:
                section_lines.append(line)
    return "".join(section_lines)


def build_release_message(context, commit: bool = False) -> str:
    # prefix = "bump: " if commit else ""
    prefix = ""
    message = f"{prefix}{PACKAGE_NICE_NAME} Release {context['next_version']}"
    message += "\n"
    message += "\n"
    message += read_changelog_section(context)
    return message


class Step(metaclass=abc.ABCMeta):
    def pre(self, context):
        pass

    def post(self, context):
        wait_for_enter()

    def run(self, context):
        try:
            self.pre(context)
            self.action(context)
            self.post(context)
        except KeyboardInterrupt:
            color_print("\nInterrupted.", color="red")
            raise SystemExit(1)

    def instruct(self, msg):
        color_print(msg, color="green")

    def print_command(self, msg):
        color_print("Run:", color="cyan", style="bright")
        color_print("\t" + msg, color="cyan", style="bright")

    def system(self, cmd: str):
        os.system(cmd)

    def execute(
        self, cmd: str, silent: bool = False, confirm: bool = True
    ) -> str:
        if not silent:
            color_print(f"Running: {cmd}", color="magenta", style="bright")
        if confirm:
            wait_for_enter()
        stdout = ""
        with subprocess.Popen(
            cmd,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE,
            text=True,
            bufsize=1,
        ) as p:
            for line in p.stdout:
                stdout += line
                if not silent:
                    print(line, end="")
        if p.returncode:
            print("--- begin stdout ---")
            print(stdout)
            print("--- end stdout ---")
            raise subprocess.CalledProcessError(
                p.returncode, p.args, stdout, p.stderr
            )
        return stdout.rstrip()

    @abc.abstractmethod
    def action(self, context):
        """Action to perform for the step"""


class GitToMaster(Step):
    def action(self, context):
        self.instruct(f"Ensuring that you're on the {BRANCH_NAME} branch")
        branch = self.execute(
            "git rev-parse --abbrev-ref HEAD", silent=True, confirm=False
        )
        if not branch == BRANCH_NAME:
            print(f"ERROR: not on {BRANCH_NAME} branch.", file=sys.stderr)
            raise SystemExit(1)

    def post(self, context):
        print("")


class UpdateChangelog(Step):
    def action(self, context):
        self.instruct(
            f"Update change log for version {context['next_version']}"
        )
        self.print_command("vi CHANGELOG.md")


class UpdateReadme(Step):
    def action(self, context):
        self.instruct("Update readme if necessary")
        self.print_command("vi README.md")


class RunTests(Step):
    def action(self, context):
        self.instruct("Running the unit tests")
        self.execute("make test")


class BumpVersionPackage(Step):
    def action(self, context):
        self.instruct("Update __version__.py with new version")
        self.system(f"vi {context['pkgname']}/__version__.py")

    def post(self, context):
        wait_for_enter()
        context["next_version"] = get_package_version(context["pkgname"])


class MakeClean(Step):
    def action(self, context):
        self.execute("make clean")


class MakeDocs(Step):
    def action(self, context):
        self.execute("make docs")


class MakeMan(Step):
    def action(self, context):
        self.execute("make man")


class InstallFromTestPyPI(Step):
    def action(self, context):
        tmpvenv = tempfile.mkdtemp(prefix="ccsv_venv_")
        parent = str(Path(tmpvenv).parent)
        self.execute(
            f"cd {parent} && python -m venv {tmpvenv} && source {tmpvenv}"
            "/bin/activate && pip install --no-cache-dir --index-url "
            "https://test.pypi.org/simple/ --extra-index-url "
            f"https://pypi.org/simple {context['pkgname']}[full]=="
            f"{context['next_version']}"
        )
        context["tmpvenv"] = tmpvenv


class TestPackage(Step):
    def action(self, context):
        self.instruct(
            f"Ensuring that the package has version {context['next_version']}"
        )
        version = self.execute(
            f"source {context['tmpvenv']}/bin/activate && clevercsv -V",
            silent=True,
            confirm=False,
        )
        if not version == context["next_version"]:
            print(
                "ERROR: version installed from TestPyPI doesn't match "
                "expected version."
            )

    def post(self, context):
        print("")


class RemoveVenv(Step):
    def action(self, context):
        self.execute(
            f"rm -rf {context['tmpvenv']}", confirm=False, silent=True
        )

    def post(self, context):
        print("")


class GitTagVersion(Step):
    def action(self, context):
        tag_message = build_release_message(context, commit=False)
        with open("./tag_message.tmp", "w") as fileobj:
            fileobj.write(tag_message)
        self.instruct("Going to tag with the following message:")
        print("--- BEGIN TAG MESSAGE ---")
        print(tag_message)
        print("--- END TAG MESSAGE ---")
        self.execute(
            f"git tag -s -F ./tag_message.tmp v{context['next_version']}"
        )
        os.unlink("./tag_message.tmp")


class GitTagPreRelease(Step):
    def action(self, context):
        last_rc_tag = get_last_release_candidate_tag(context["next_version"])
        rc_count = (
            1 if last_rc_tag is None else int(last_rc_tag.split(".")[-1]) + 1
        )
        context["rc_count"] = rc_count
        self.instruct("Tagging version as a pre-release")
        self.execute(
            f'git tag -s -m "{PACKAGE_NICE_NAME} Release v'
            f"{context['next_version']} (release candidate {rc_count})\" v"
            f"{context['next_version']}-rc.{rc_count}"
        )


class GitAdd(Step):
    def action(self, context):
        self.instruct("Commit any final changes for release to git")
        self.print_command("git gui")


class GitAddVersionAndMan(Step):
    def action(self, context):
        self.instruct("Add version and man pages to git and commit")
        self.execute(
            f"git add {context['pkgname']}/__version__.py", confirm=False
        )
        self.execute(f"git add {MAN_DIRECTORY}", confirm=False)
        # TODO: Fail gracefully if `git diff --cached --exit-code` is 0
        self.execute("git commit -m 'bump: update version and manpages'")


class GitAddRelease(Step):
    def action(self, context):
        self.instruct("Add Changelog & Readme to git")
        self.execute("git add CHANGELOG.md", confirm=False)
        self.execute("git add README.md", confirm=False)
        self.execute("git add docs/CHANGELOG.rst", confirm=False)
        self.execute("git add man/", confirm=False)
        self.instruct("Going to commit with the following message:")

        commit_message = build_release_message(context, commit=True)
        print("--- BEGIN COMMIT MESSAGE ---")
        print(commit_message)
        print("--- END COMMIT MESSAGE ---")
        with open("./commit_message.tmp", "w") as fileobj:
            fileobj.write(commit_message)
        self.execute("git commit --file ./commit_message.tmp")
        os.unlink("./commit_message.tmp")


class PushToGitHub(Step):
    def action(self, context):
        self.execute(f"git push -u --tags origin {BRANCH_NAME}")


class WaitForCI(Step):
    def action(self, context):
        webbrowser.open(URLS["CI"])
        self.instruct("Wait for CI to complete and verify that its successful")


class WaitForRTD(Step):
    def action(self, context):
        webbrowser.open(URLS["RTD"])
        self.instruct(
            "Wait for ReadTheDocs to complete and verify that its successful"
        )


class GitHubRelease(Step):
    def action(self, context):
        webbrowser.open(URLS["tags"])
        self.instruct("Create release from tag and embed release notes")


class UpdatePreCommitDummy(Step):
    def action(self, context):
        self.instruct(
            f"Update the pre-commit dummy package ({URLS['dummy']}) "
            "by running ``make release`` there"
        )


def main(target=None):
    colorama.init()
    procedure = [
        ("gittomaster", GitToMaster()),
        ("clean1", MakeClean()),
        # ("docs1", MakeDocs()),
        # ("man1", MakeMan()),
        ("runtests", RunTests()),
        # trigger CI to run tests on all platforms
        ("push1", PushToGitHub()),
        ("ci1", WaitForCI()),
        ("waitrtd", WaitForRTD()),
        ("bumpversion", BumpVersionPackage()),
        ("gitadd2", GitAdd()),
        ("gittagpre", GitTagPreRelease()),
        # trigger CI to run tests using cibuildwheel
        ("push2", PushToGitHub()),
        ("ci2", WaitForCI()),
        ("changelog", UpdateChangelog()),
        ("readme", UpdateReadme()),
        ("clean2", MakeClean()),
        ("docs2", MakeDocs()),
        ("man2", MakeMan()),
        ("install", InstallFromTestPyPI()),
        ("testpkg", TestPackage()),
        ("remove_venv", RemoveVenv()),
        ("addrelease", GitAddRelease()),
        ("tagfinal", GitTagVersion()),
        # triggers Travis to build with cibw and push to PyPI
        ("push3", PushToGitHub()),
        ("ci3", WaitForCI()),
        ("gh_release", GitHubRelease()),
        ("pre-commit", UpdatePreCommitDummy()),
    ]
    context = {}
    context["pkgname"] = get_package_name()
    context["prev_version_tag"] = get_last_version_tag()
    context["next_version"] = get_package_version(context["pkgname"])
    skip = True if target else False
    for name, step in procedure:
        if not name == target and skip:
            continue
        skip = False
        step.run(context)
    color_print("\nDone!", color="yellow", style="bright")


if __name__ == "__main__":
    target = sys.argv[1] if len(sys.argv) > 1 else None
    main(target=target)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Drop-in replacement for the Python csv writer class.

Author: Gertjan van den Burg

"""

from __future__ import annotations

import csv

from typing import TYPE_CHECKING
from typing import Any
from typing import Iterable

import _csv

if TYPE_CHECKING:
    from clevercsv._types import SupportsWrite

from clevercsv._types import _DialectLike

from .dialect import SimpleDialect
from .exceptions import Error

DIALECT_KEYS = [
    "skipinitialspace",
    "doublequote",
    "strict",
    "delimiter",
    "escapechar",
    "lineterminator",
    "quotechar",
    "quoting",
]


class writer:
    def __init__(
        self,
        csvfile: SupportsWrite[str],
        dialect: _DialectLike = "excel",
        **fmtparams: Any,
    ) -> None:
        self.original_dialect = dialect
        self.dialect: type[_csv.Dialect] = self._make_python_dialect(
            dialect, **fmtparams
        )
        self._writer = _csv.writer(csvfile, dialect=self.dialect)

    def _make_python_dialect(
        self, dialect: _DialectLike, **fmtparams: Any
    ) -> type[_csv.Dialect]:
        d: _DialectLike = ""
        if isinstance(dialect, str):
            d = _csv.get_dialect(dialect)
        elif isinstance(dialect, _csv.Dialect):
            d = dialect
        elif isinstance(dialect, SimpleDialect):
            d = dialect.to_csv_dialect()
        elif dialect in [csv.excel, csv.excel_tab, csv.unix_dialect]:
            d = dialect
        else:
            raise ValueError(f"Unknown dialect type: {dialect}")

        # Override properties from format parameters
        props = {k: getattr(d, k) for k in DIALECT_KEYS if hasattr(d, k)}
        for key, value in fmtparams.items():
            props[key] = value

        # lineterminator must be set
        if "lineterminator" not in props or props["lineterminator"] is None:
            props["lineterminator"] = "\n"

        # We have to subclass the csv.Dialect
        newdialect = type("dialect", (csv.Dialect,), props)
        return newdialect

    def writerow(self, row: Iterable[Any]) -> Any:
        try:
            return self._writer.writerow(row)
        except csv.Error as e:
            raise Error(str(e))

    def writerows(self, rows: Iterable[Iterable[Any]]) -> Any:
        try:
            return self._writer.writerows(rows)
        except csv.Error as e:
            raise Error(str(e))



from typing import Any

CODE_SIZE: int
MAGIC: int
copyright: str

def compile(*args, **kwargs) -> Any: ...
def fold_case(*args, **kwargs) -> Any: ...
def get_all_cases(*args, **kwargs) -> Any: ...
def get_code_size(*args, **kwargs) -> Any: ...
def get_expand_on_folding(*args, **kwargs) -> Any: ...
def get_properties(*args, **kwargs) -> Any: ...
def has_property_value(*args, **kwargs) -> Any: ...



# -*- coding: utf-8 -*-

"""
Detect the dialect with very strict functional tests.

This module uses so-called "normal forms" to detect the dialect of CSV files.
Normal forms are detected with strict functional tests. The normal forms are
used as a pre-test to check if files are simple enough that computing the data
consistency measure is not necessary.

Author: Gertjan van den Burg

"""

import itertools

from typing import Callable
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple

import regex

from .dialect import SimpleDialect
from .escape import is_potential_escapechar
from .utils import pairwise

DELIMS: List[str] = [",", ";", "|", "\t"]
QUOTECHARS: List[str] = ["'", '"']


def detect_dialect_normal(
    data: str,
    encoding: str = "UTF-8",
    delimiters: Optional[Iterable[str]] = None,
    verbose: bool = False,
) -> Optional[SimpleDialect]:
    """Detect the normal form of a file from a given sample

    Parameters
    ----------
    data : str
        The data as a single string

    encoding : str
        The encoding of the data

    Returns
    -------
    dialect : SimpleDialect
        The dialect detected using normal forms, or None if no such dialect can
        be found.
    """
    if delimiters is None:
        delimiters = DELIMS
    delimiters = list(delimiters)
    for delim, quotechar in itertools.product(delimiters, QUOTECHARS):
        if maybe_has_escapechar(data, encoding, delim, quotechar):
            if verbose:
                print("Not normal, has potential escapechar.")
            return None

    form_and_dialect: List[
        Tuple[int, Callable[[List[str], SimpleDialect], bool], SimpleDialect]
    ] = []

    for delim in delimiters:
        dialect = SimpleDialect(delimiter=delim, quotechar="", escapechar="")
        form_and_dialect.append((2, is_form_2, dialect))

    for delim, quotechar in itertools.product(delimiters, QUOTECHARS):
        dialect = SimpleDialect(
            delimiter=delim, quotechar=quotechar, escapechar=""
        )
        form_and_dialect.append((1, is_form_1, dialect))
        form_and_dialect.append((3, is_form_3, dialect))
        form_and_dialect.append((5, is_form_5, dialect))
    for quotechar in QUOTECHARS:
        dialect = SimpleDialect(
            delimiter="", quotechar=quotechar, escapechar=""
        )
        form_and_dialect.append((4, is_form_4, dialect))

    form_and_dialect.append(
        (
            4,
            is_form_4,
            SimpleDialect(delimiter="", quotechar="", escapechar=""),
        )
    )

    rows = split_file(data)

    for ID, form_func, dialect in form_and_dialect:
        if form_func(rows, dialect):
            if verbose:
                print("Matched normal form %i." % ID)
            return dialect
    if verbose:
        print("Didn't match any normal forms.")
    return None


def is_quoted_cell(cell: str, quotechar: str) -> bool:
    if len(cell) < 2:
        return False
    return cell[0] == quotechar and cell[-1] == quotechar


def is_any_quoted_cell(cell: str) -> bool:
    return is_quoted_cell(cell, "'") or is_quoted_cell(cell, '"')


def is_any_partial_quoted_cell(cell: str) -> bool:
    if len(cell) < 1:
        return False
    return (
        cell[0] == '"' or cell[0] == "'" or cell[-1] == '"' or cell[-1] == "'"
    )


def is_empty_quoted(cell: str, quotechar: str) -> bool:
    return len(cell) == 2 and is_quoted_cell(cell, quotechar)


def is_empty_unquoted(cell: str) -> bool:
    return cell == ""


def is_any_empty(cell: str) -> bool:
    return (
        is_empty_unquoted(cell)
        or is_empty_quoted(cell, "'")
        or is_empty_quoted(cell, '"')
    )


def has_delimiter(string: str, delim: str) -> bool:
    return delim in string


def has_nested_quotes(string: str, quotechar: str) -> bool:
    return quotechar in string[1:-1]


def maybe_has_escapechar(
    data: str, encoding: str, delim: str, quotechar: str
) -> bool:
    if delim not in data and quotechar not in data:
        return False
    for u, v in pairwise(data):
        if v in [delim, quotechar] and is_potential_escapechar(u, encoding):
            return True
    return False


def strip_trailing_crnl(data: str) -> str:
    while data.endswith("\n"):
        data = data.rstrip("\n")
    while data.endswith("\r"):
        data = data.rstrip("\r")
    return data


def every_row_has_delim(rows: List[str], dialect: SimpleDialect) -> bool:
    assert dialect.delimiter is not None
    for row in rows:
        if not has_delimiter(row, dialect.delimiter):
            return False
    return True


def every_row_has_delim_and_is_the_same_length(
    rows: List[str], dialect: SimpleDialect
) -> bool:
    assert dialect.delimiter is not None
    if len(rows) == 0:
        return False

    first_len = len(split_row(rows[0], dialect))
    for row in rows:
        if not has_delimiter(row, dialect.delimiter):
            return False
        if len(split_row(row, dialect)) != first_len:
            return False
    return True


def is_elementary(cell: str) -> bool:
    return (
        regex.fullmatch(r"[a-zA-Z0-9\.\_\&\-\@\+\%\(\)\ \/]+", cell)
        is not None
    )


def split_file(data: str) -> List[str]:
    data = strip_trailing_crnl(data)
    if "\r\n" in data:
        return data.split("\r\n")
    elif "\n" in data:
        return data.split("\n")
    elif "\r" in data:
        return data.split("\r")
    return [data]


def split_row(row: str, dialect: SimpleDialect) -> List[str]:
    # no nested quotes
    assert dialect.quotechar is not None
    if (not dialect.quotechar) or (dialect.quotechar not in row):
        if not dialect.delimiter:
            return [row]
        return row.split(dialect.delimiter)

    cells = []
    current_cell = ""
    in_quotes = False
    for c in row:
        if c == dialect.delimiter and not in_quotes:
            cells.append(current_cell)
            current_cell = ""
        elif c == dialect.quotechar:
            in_quotes = not in_quotes
            current_cell += c
        else:
            current_cell += c
    if current_cell:
        cells.append(current_cell)
    return cells


def is_form_1(rows: List[str], dialect: SimpleDialect) -> bool:
    # All cells quoted, quoted empty allowed, no nested quotes, more than one
    # column
    assert dialect.quotechar is not None

    if not every_row_has_delim_and_is_the_same_length(rows, dialect):
        return False

    for row in rows:
        cells = split_row(row, dialect)
        if len(cells) == 1:
            return False
        for cell in cells:
            # No empty cells
            if is_empty_unquoted(cell):
                return False

            # All cells must be quoted
            if not is_quoted_cell(cell, dialect.quotechar):
                return False

            # No quotes inside quotes
            if has_nested_quotes(cell, dialect.quotechar):
                return False

    return True


def is_form_2(rows: List[str], dialect: SimpleDialect) -> bool:
    # All unquoted, empty allowed, all elementary
    if not every_row_has_delim_and_is_the_same_length(rows, dialect):
        return False

    for row in rows:
        cells = split_row(row, dialect)
        if len(cells) == 1:
            return False
        for cell in cells:
            # All cells must be unquoted
            if is_any_quoted_cell(cell):
                return False
            # All cells must not be partially quoted
            if is_any_partial_quoted_cell(cell):
                return False
            # Cells have to be elementary
            if not is_empty_unquoted(cell) and not is_elementary(cell):
                return False
    return True


def is_form_3(rows: List[str], dialect: SimpleDialect) -> bool:
    # some quoted, some not quoted, no empty, no nested quotes
    assert dialect.quotechar is not None

    if not every_row_has_delim_and_is_the_same_length(rows, dialect):
        return False
    if len(rows) <= 1:
        return False

    for row in rows:
        cells = split_row(row, dialect)
        if len(cells) == 1:
            return False
        for cell in cells:
            if is_any_empty(cell):
                return False

            # if it is quoted
            if is_any_quoted_cell(cell):
                # but not quoted with the quotechar of the dialect
                if not is_quoted_cell(cell, dialect.quotechar):
                    # then this form isn't right
                    return False
            # if it's not quoted
            else:
                # and it's not elementary
                if not is_elementary(cell):
                    # then this form isn't right
                    return False

    return True


def is_form_4(rows: List[str], dialect: SimpleDialect) -> bool:
    # no delim, single column (either entirely quoted or entirely unquoted)
    assert dialect.quotechar is not None

    if len(rows) <= 1:
        return False

    unquoted_search = regex.compile(r"[^A-Za-z0-9.\_&\-]").search
    quoted_search = regex.compile(r"[^A-Za-z0-9.\_&\-\ ]").search
    for row in rows:
        cell = row[:]
        if dialect.quotechar == "":
            if is_any_quoted_cell(cell):
                return False
            if unquoted_search(cell):
                return False
        else:
            if not is_quoted_cell(cell, dialect.quotechar):
                return False
            if quoted_search(cell[1:-1]):
                return False

    return True


def is_form_5(rows: List[str], dialect: SimpleDialect) -> bool:
    # all rows quoted, no nested quotes
    # basically form 2 but with quotes around each row
    if not every_row_has_delim(rows, dialect):
        return False
    if len(rows) <= 1:
        return False

    for row in rows:
        if not (
            len(row) > 2
            and row[0] == dialect.quotechar
            and row[-1] == dialect.quotechar
        ):
            return False

    newrows = []
    for row in rows:
        newrows.append(row[1:-1])

    return is_form_2(newrows, dialect)



from typing import Any

from regex._regex_core import VERSION0

def match(
    pattern,
    string,
    flags: int = ...,
    pos: Any | None = ...,
    endpos: Any | None = ...,
    partial: bool = ...,
    concurrent: Any | None = ...,
    timeout: Any | None = ...,
    ignore_unused: bool = ...,
    **kwargs
): ...
def fullmatch(
    pattern,
    string,
    flags: int = ...,
    pos: Any | None = ...,
    endpos: Any | None = ...,
    partial: bool = ...,
    concurrent: Any | None = ...,
    timeout: Any | None = ...,
    ignore_unused: bool = ...,
    **kwargs
): ...
def search(
    pattern,
    string,
    flags: int = ...,
    pos: Any | None = ...,
    endpos: Any | None = ...,
    partial: bool = ...,
    concurrent: Any | None = ...,
    timeout: Any | None = ...,
    ignore_unused: bool = ...,
    **kwargs
): ...
def sub(
    pattern,
    repl,
    string,
    count: int = ...,
    flags: int = ...,
    pos: Any | None = ...,
    endpos: Any | None = ...,
    concurrent: Any | None = ...,
    timeout: Any | None = ...,
    ignore_unused: bool = ...,
    **kwargs
): ...
def subf(
    pattern,
    format,
    string,
    count: int = ...,
    flags: int = ...,
    pos: Any | None = ...,
    endpos: Any | None = ...,
    concurrent: Any | None = ...,
    timeout: Any | None = ...,
    ignore_unused: bool = ...,
    **kwargs
): ...
def subn(
    pattern,
    repl,
    string,
    count: int = ...,
    flags: int = ...,
    pos: Any | None = ...,
    endpos: Any | None = ...,
    concurrent: Any | None = ...,
    timeout: Any | None = ...,
    ignore_unused: bool = ...,
    **kwargs
): ...
def subfn(
    pattern,
    format,
    string,
    count: int = ...,
    flags: int = ...,
    pos: Any | None = ...,
    endpos: Any | None = ...,
    concurrent: Any | None = ...,
    timeout: Any | None = ...,
    ignore_unused: bool = ...,
    **kwargs
): ...
def split(
    pattern,
    string,
    maxsplit: int = ...,
    flags: int = ...,
    concurrent: Any | None = ...,
    timeout: Any | None = ...,
    ignore_unused: bool = ...,
    **kwargs
): ...
def splititer(
    pattern,
    string,
    maxsplit: int = ...,
    flags: int = ...,
    concurrent: Any | None = ...,
    timeout: Any | None = ...,
    ignore_unused: bool = ...,
    **kwargs
): ...
def findall(
    pattern,
    string,
    flags: int = ...,
    pos: Any | None = ...,
    endpos: Any | None = ...,
    overlapped: bool = ...,
    concurrent: Any | None = ...,
    timeout: Any | None = ...,
    ignore_unused: bool = ...,
    **kwargs
): ...
def finditer(
    pattern,
    string,
    flags: int = ...,
    pos: Any | None = ...,
    endpos: Any | None = ...,
    overlapped: bool = ...,
    partial: bool = ...,
    concurrent: Any | None = ...,
    timeout: Any | None = ...,
    ignore_unused: bool = ...,
    **kwargs
): ...
def compile(
    pattern, flags: int = ..., ignore_unused: bool = ..., **kwargs
): ...
def purge() -> None: ...
def cache_all(value: bool = ...): ...
def template(pattern, flags: int = ...): ...
def escape(pattern, special_only: bool = ..., literal_spaces: bool = ...): ...

DEFAULT_VERSION = VERSION0
Pattern: Any
Match: Any
Regex = compile

# Names in __all__ with no definition:
#   A
#   ASCII
#   B
#   BESTMATCH
#   D
#   DEBUG
#   DOTALL
#   E
#   ENHANCEMATCH
#   F
#   FULLCASE
#   I
#   IGNORECASE
#   L
#   LOCALE
#   M
#   MULTILINE
#   P
#   POSIX
#   R
#   REVERSE
#   S
#   Scanner
#   T
#   TEMPLATE
#   U
#   UNICODE
#   V0
#   V1
#   VERBOSE
#   VERSION0
#   VERSION1
#   W
#   WORD
#   X
#   __doc__
#   __version__
#   error
